"""This module contains the SmartTest class which you can use to communicate with the leak tester."""
import serial
import logging
import argparse
import time
from datetime import datetime

from qutil.io import CsvLogger

__all__ = ["SmartTest"]

logger = logging.getLogger('smart_test')


class CheckSumError(ValueError):
    pass


def calc_checksum(msg):
    return b'%03i' % (sum(msg)%256)


def to_valid_message(msg: bytes):
    return msg + calc_checksum(msg) + b'\r'


def parse_data(data: bytes, fmt: str):
    if fmt in ('boolean_old', 0):
        assert data in (b'0'*6, b'1'*6)
        return bool(int(data))
    
    if fmt in ('u_integer', 1):
        assert len(data) == 6
        return int(data)
    
    if fmt in ('u_real', 2):
        assert len(data) == 6
        return int(data) / 100
    
    if fmt in ('string', 4):
        assert len(data) == 6
        return data.decode('ascii')
    
    if fmt in ('boolean_new', 6):
        assert len(data) == 1
        return bool(int(data))
    
    if fmt in ('u_short_int', 7):
        assert len(data) == 3
        return int(data)
    
    if fmt in ('u_expo_new', 10):
        assert len(data) == 6
        mantissa = data[:4].decode()
        exponent = str(int(data[4:]) - 20 - 3)
        # use parsing logic for most exact result
        return float('%sE%s' % (mantissa, exponent))
        
    if fmt in ('string16', 11):
        assert len(data) == 16
        return data.decode()
    
    raise KeyError('unknown format argument for parsing', data, fmt)


def encode_data(value, fmt: str) -> bytes:
    if fmt in ('boolean_old', 0):
        assert value in (True, False, 0, 1)
        return 6*b'1' if value else 6*b'0'
    
    if fmt in ('u_integer', 1):
        assert value < 1000000 and value >= 0
        return b'%06u' % value
    
    if fmt in ('u_real', 2):
        assert value >= 0 and value < 10000
        return (b'%07.2f' % value).replace(b'.', b'')
    
    if fmt in ('string', 4):
        assert len(value) == 4
        return value.encode('ascii')
    
    if fmt in ('boolean_new', 6):
        assert value in (True, False, 0, 1)
        return b'1' if value else b'0'
    
    if fmt in ('u_short_int', 7):
        assert value >= 0 and value < 1000
        return b'%03u' % value
    
    if fmt in ('u_expo_new', 10):
        assert value >= 0
        str_repr = (b'%.3E' % value)
        mantissa, exponent = str_repr.replace(b'.', b'').split(b'E')
        exponent = int(exponent) + 20
        assert exponent >= 0 and exponent < 100
        
        return mantissa + b'%02u' % exponent
        
    if fmt in ('string16', 11):
        assert len(value) == 16
        return value.encode('ascii')
    
    raise KeyError('unknown format argument for parsing', value, fmt)
    

def parse_msg(msg: bytes, fmt: str):
    addr = msg[:3]
    logger.debug('parsing message from %r', addr)
    
    action = msg[3:5]
    param_num = int(msg[5:8])
    data_len = int(msg[8:10])
    
    data = msg[10:10+data_len]
    
    expected_check_sum = calc_checksum(msg[:10+data_len])
    
    check_sum = msg[10+data_len:10+data_len+3]
    
    if expected_check_sum != check_sum:
        raise CheckSumError(msg)
    return action, param_num, parse_data(data, fmt)


class SmartTest:
    """Talk to PfeifferVacuum leak tester via serial interface (USB serial adapter).
    
    Currently only get_leakrate_mbarls is implemented as a method but you can easily extend this to other get methods using
    the parameter number and the value format from the manual. For setting something you need to implement set method with the action b'10'
    """
    def __init__(self, device: str, address: int):
        assert address < 1000
        self.device = device
        self.address = b'%03i' % address

    def query(self, param_number: int, fmt):
        param_b = b'%03i' % param_number
        
        msg = self.address + b'00' + param_b + b'02' + b'=?'
        query = to_valid_message(msg)
        
        with serial.Serial(self.device) as ser:
            ser.write(query)
            answer = ser.read_until(b'\r')
        
        action, param_num, value = parse_msg(answer, fmt)
        logger.info('received: %r', (action, param_num, value))
        assert action == b'10'
        assert param_num == param_number, "%u != %u" % (param_num, param_number)
        return value
    
    def get_leakrate_mbarls(self) -> float:
        return self.query(670, 'u_expo_new')


def periodically_log_leak_rate(filename, period, device, address):
    leak_detector = SmartTest(device, address)
    csv_logger = CsvLogger(filename, ['Time', 'Leakrate(mbar x l/s)'])

    while True:
        leakrate = leak_detector.get_leakrate_mbarls()
        current_time = datetime.now()

        csv_logger.write('%s' % current_time, '%.4e' % leakrate)
    
        print('Current leakrate %.4e pausing for %f sec.' % (leakrate, period), end='\r')
        time.sleep(period)


default_prefix = '%Y_%m_%d_%H_%M_%S_'

parser = argparse.ArgumentParser(description='Periodically log the leak rate to the given file as tab-separated values.',
                                 formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('--filename', default='smart_test_fast.log', help='Filename of log file. By default prefixed by a date/time string.')
parser.add_argument('--period', default=1., type=float, help='Wait time between read-outs')
parser.add_argument('--device', default='/dev/ttyUSB0', help='Device path')
parser.add_argument('--address', default=1, type=int, help='See SmartTester settings')
parser.add_argument('--no-date-prefix', action='store_true', help='Omit the default %s prefix' % default_prefix.replace('%', '%%'))


def main():
    args = vars(parser.parse_args())

    filename = args.pop('filename')
    if not args.pop('no_date_prefix'):
        filename = datetime.now().strftime(default_prefix) + filename
    
    periodically_log_leak_rate(filename=filename, **args)


if __name__ == '__main__':
    main()



"""This module contains utility functions for linear algebra manipulations."""
import operator
import string
import sys
import warnings
from functools import reduce
from itertools import zip_longest
from typing import List, NamedTuple, Sequence, Tuple, Union

import numpy as np
from numpy import linalg, ndarray
from scipy.linalg import polar

if sys.version_info >= (3, 13):
    from warnings import deprecated
else:
    from typing_extensions import deprecated

from .math import abs2, cexp, remove_float_errors, max_abs_diff, max_rel_diff

try:
    cexp = deprecated('Use math.cexp instead')(cexp)
except AttributeError:
    _cexp = cexp

    def cexp(*args, **kwargs):
        warnings.warn('Use math.cexp instead', DeprecationWarning)
        return _cexp(*args, **kwargs)
try:
    abs2 = deprecated('Use math.abs2 instead')(abs2)
except AttributeError:
    _abs2 = abs2

    def abs2(*args, **kwargs):
        warnings.warn('Use math.abs2 instead', DeprecationWarning)
        return _abs2(*args, **kwargs)
remove_float_errors = deprecated('Use math.remove_float_errors instead')(remove_float_errors)
max_abs_diff = deprecated('Use math.max_abs_diff instead')(max_abs_diff)
max_rel_diff = deprecated('Use math.max_rel_diff instead')(max_rel_diff)

__all__ = ['check_phase_eq', 'density', 'dot_HS', 'mdot', 'pauli_expm', 'ptrace', 'sparsity',
           'tensor', 'paulis']


class PauliMatrices(NamedTuple):
    """Collection of pauli matrices that can be addressed by 'name' or by index.

    Examples
    --------
    >>> from qutil.linalg import paulis
    >>> paulis.sigma_z
    array([[ 1,  0],
           [ 0, -1]])
    >>> import numpy as np
    >>> np.testing.assert_equal(paulis.sigma_0 @ paulis.sigma_x, paulis.sigma_x)
    >>> np.testing.assert_equal(paulis.sigma_x @ paulis.sigma_x, paulis.sigma_0)
    >>> np.testing.assert_equal(paulis.sigma_x @ paulis.sigma_y, 1j * paulis.sigma_z)
    >>> np.testing.assert_equal(paulis.sigma_y.transpose().conjugate(), paulis.sigma_y)
    """
    sigma_0: np.ndarray = np.array([[1, 0], [0, 1]])
    sigma_x: np.ndarray = np.array([[0, 1], [1, 0]])
    sigma_y: np.ndarray = np.array([[0, -1j], [1j, 0]])
    sigma_z: np.ndarray = np.array([[1, 0], [0, -1]])


paulis = PauliMatrices()


def closest_unitary(Q: ndarray, subspace: Sequence[int] = None) -> ndarray:
    """Compute the closest unitary to ``Q[np.ix_(*subspace)]`` on a given
    subspace using left polar decomposition."""
    if subspace is None:
        subspace = (range(Q.shape[-2]), range(Q.shape[-1]))

    idx = np.ix_(*subspace)
    V = Q.copy()
    if Q.ndim == 2:
        V[idx], _ = polar(Q[idx], side='left')
    else:
        for i, q in enumerate(Q):
            V[i][idx], _ = polar(q[idx], side='left')

    return V


def tensor(*args, rank: int = 2,
           optimize: Union[bool, str] = False) -> ndarray:
    """
    Fast, flexible tensor product using einsum. The product is taken over the
    last *rank* axes and broadcast over the remaining axes which thus need to
    follow numpy broadcasting rules. Note that vectors are treated as rank 2
    tensors with shape (1, x) or (x, 1).

    For example, the following shapes are compatible:

     - ``rank == 2`` (e.g. matrices or vectors)::

        (a, b, c, d, d), (a, b, c, e, e) -> (a, b, c, d*e, d*e)
        (a, b, c), (a, d, e) -> (a, b*d, c*e)
        (a, b), (c, d, e) -> (c, a*d, b*e)
        (1, a), (b, 1, c) -> (b, 1, a*c)
     - ``rank == 1``::

        (a, b), (a, c) -> (a, b*c)
        (a, b, 1), (a, c) -> (a, b, c)

    Parameters
    ----------
    args : array_like
        The elements of the tensor product
    rank : int, optional (default: 2)
        The rank of the tensors. E.g., for a Kronecker product between two
        matrices ``rank == 2``. The remaining axes are broadcast over.
    optimize : bool|str, optional (default: False)
        Optimize the tensor contraction order. Passed through to
        :meth:`numpy.einsum`.

    Examples
    --------
    >>> Z = np.diag([1, -1])
    >>> np.array_equal(tensor(Z, Z), np.kron(Z, Z))
    True

    >>> A, B = np.arange(2), np.arange(2, 5)
    >>> tensor(A, B, rank=1)
    array([0, 0, 0, 2, 3, 4])

    >>> args = np.random.randn(4, 10, 3, 2)
    >>> result = tensor(*args, rank=1)
    >>> result.shape == (10, 3, 2**4)
    True
    >>> result = tensor(*args, rank=2)
    >>> result.shape == (10, 3**4, 2**4)
    True

    >>> A, B = np.random.randn(1, 3), np.random.randn(3, 4)
    >>> result = tensor(A, B)
    >>> result.shape == (1*3, 3*4)
    True

    >>> A, B = np.random.randn(3, 1, 2), np.random.randn(2, 2, 2)
    >>> try:
    ...     result = tensor(A, B, rank=2)
    ... except ValueError as err:  # cannot broadcast over axis 0
    ...     print(err)
    Incompatible shapes (3, 1, 2) and (2, 2, 2) for tensor product of rank 2.
    >>> result = tensor(A, B, rank=3)
    >>> result.shape == (3*2, 1*2, 2*2)
    True

    See Also
    --------
    :meth:`numpy.kron`
    """
    chars = string.ascii_letters
    # All the subscripts we need
    A_chars = chars[:rank]
    B_chars = chars[rank:2 * rank]
    subscripts = '...{},...{}->...{}'.format(
        A_chars, B_chars, ''.join(i + j for i, j in zip(A_chars, B_chars))
    )

    def _tensor_product_shape(shape_A: Sequence[int], shape_B: Sequence[int],
                              rank: int):
        """Get shape of the tensor product between A and B of rank rank"""
        broadcast_shape = ()
        # Loop over dimensions from last to first, filling the 'shorter' shape
        # with 1's once it is exhausted
        for dims in zip_longest(shape_A[-rank - 1::-1], shape_B[-rank - 1::-1],
                                fillvalue=1):
            if 1 in dims:
                # Broadcast 1-d of argument to dimension of other
                broadcast_shape = (max(dims),) + broadcast_shape
            elif len(set(dims)) == 1:
                # Both arguments have same dimension on axis.
                broadcast_shape = dims[:1] + broadcast_shape
            else:
                raise ValueError('Incompatible shapes ' +
                                 '{} and {} '.format(shape_A, shape_B) +
                                 'for tensor product of rank {}.'.format(rank))

        # Shape of the actual tensor product is product of each dimension,
        # again broadcasting if need be
        tensor_shape = tuple(
            reduce(operator.mul, dimensions) for dimensions in zip_longest(
                shape_A[:-rank - 1:-1], shape_B[:-rank - 1:-1], fillvalue=1
            )
        )[::-1]

        return broadcast_shape + tensor_shape

    def binary_tensor(A, B):
        """Compute the Kronecker product of two tensors"""
        # Add dimensions so that each arg has at least ndim == rank
        while A.ndim < rank:
            A = A[None, :]

        while B.ndim < rank:
            B = B[None, :]

        outshape = _tensor_product_shape(A.shape, B.shape, rank)
        return np.einsum(subscripts, A, B, optimize=optimize).reshape(outshape)

    # Compute the tensor products in a binary tree-like structure, calculating
    # the product of two leaves and working up. This is more memory-efficient
    # than reduce(binary_tensor, args) which computes the products
    # left-to-right.
    n = len(args)
    bit = n % 2
    while n > 1:
        args = args[:bit] + tuple(binary_tensor(*args[i:i + 2])
                                  for i in range(bit, n, 2))

        n = len(args)
        bit = n % 2

    return args[0]


def mdot(arr: Sequence, axis: int = 0) -> ndarray:
    """Multiple matrix products along axis"""
    return reduce(np.matmul, arr.swapaxes(0, axis))


def ptrace(arr: ndarray, dims: Sequence[int], system: int = 1) -> ndarray:
    r"""
    Partial trace of bipartite matrix.

    Parameters
    ----------
    arr : ndarray
        Matrix of bipartite system on the last two axes. There are no
        constraints on the shape of the other axes.

        .. math::

            \verb|arr|\equiv A\otimes B
    dims : Sequence of ints
        The dimensions of system A and B
    system : int
        The system to trace out, either 0 (A) or 1 (B)

    Examples
    --------
    >>> d_A, d_B = 3, 2
    >>> A = np.arange(d_A**2).reshape(1, d_A, d_A)
    >>> B = np.arange(d_B**2).reshape(1, d_B, d_B)
    >>> C = tensor(A, B)
    >>> ptrace(C, (d_A, d_B), 0)
    array([[[ 0, 12],
            [24, 36]]])
    >>> ptrace(C, (d_A, d_B), 1)
    array([[[ 0,  3,  6],
            [ 9, 12, 15],
            [18, 21, 24]]])

    """
    if system == 0:
        einsum_str = '...ijik'
    elif system == 1:
        einsum_str = '...ikjk'

    return np.einsum(einsum_str, arr.reshape(*arr.shape[:-2], *dims, *dims))


def pauli_expm(a: Sequence, nonzero_ind: List[bool]) -> ndarray:
    r"""
    Faster method of calculating a Pauli exponential

    .. math::
        \exp(-i(\vec{a}\cdot\vec{\sigma})) = I\cos|\vec{a}|
            - i(\vec{a}\cdot\vec{\sigma})\sin|\vec{a}|

    than :meth:`scipy.linalg.expm`.

    Parameters
    ----------
    a : array_like, shape (n, ...)
        Array with 0 < n <= 3 cartesian components on the first axis.
    nonzero_ind : list of three bools
        List of three Booleans indicating those cartesian components that are
        nonzero, i.e., which of the three (x, y, z) are the first axis of *a*.
        Accordingly, ``sum(nonzero_ind) == a.shape[0]``.

    Examples
    --------
    A Hadamard gate, i.e. a rotation by pi about X + Z (dividing by sqrt(2)
    normalizes the rotation axis vector):

    >>> a = np.full((2,), np.pi/2/np.sqrt(2))
    >>> pauli_expm(a, [True, False, True])  # a[0]=a[2]=pi/2/sqrt(2), a[1]=0
    array([[6.123234e-17-0.70710678j, 0.000000e+00-0.70710678j],
           [0.000000e+00-0.70710678j, 6.123234e-17+0.70710678j]])
    """
    a = np.asarray(a)
    idx = np.asarray(nonzero_ind)
    if not sum(idx) == a.shape[0]:
        raise ValueError('nonzero_ind should have the same number of True '
                         f'entries as a.shape[0] = {a.shape[0]}!')

    p0, pi = np.split(paulis, [1])
    # Twice the rotation angle
    theta = linalg.norm(a, axis=0)
    # Rotation axis
    n = a / theta
    # Replace nans originating from divide by zero
    n[np.isnan(n)] = 0
    real = np.einsum('ijk,...->jk...', p0, np.cos(theta))
    imag = np.einsum('i...,ijk,...->jk...', n, pi[idx], np.sin(theta))
    result = real - 1j * imag
    return result.transpose([2, 0, 1]) if result.ndim == 3 else result


def check_phase_eq(psi: Sequence,
                   phi: Sequence,
                   eps: float = None,
                   normalized: bool = False) -> Tuple[bool, float]:
    r"""
    Checks whether psi and phi are equal up to a global phase, i.e.

    .. math::
        |\psi\rangle = e^{i\chi}|\phi\rangle \Leftrightarrow
        \langle \phi|\psi\rangle = e^{i\chi},

    and returns the phase. If the first return value is false, the second is
    meaningless in this context. psi and phi can also be operators.

    Parameters
    ----------
    psi, phi : array_like
        Vectors or operators to be compared
    eps : float
        The tolerance below which the two objects are treated as equal, i.e.,
        the function returns ``True`` if ``abs(1 - modulus) <= eps``.
    normalized : bool
        Flag indicating if *psi* and *phi* are normalized with respect to the
        Hilbert-Schmidt inner product :meth:`dot_HS`.

    Examples
    --------
    >>> import qutil.qi
    >>> psi = qutil.qi.paulis[1]
    >>> phi = qutil.qi.paulis[1]*np.exp(1j*1.2345)
    >>> eq, phase = check_phase_eq(psi, phi)
    >>> print(eq)
    True
    >>> print(phase)
    1.2345000000000603
    """
    if eps is None:
        # Tolerance the floating point eps times the # of flops for the matrix
        # multiplication, i.e. for psi and phi n x m matrices 2*n**2*m
        try:
            psi_eps = np.finfo(psi.dtype).eps
        except ValueError:
            # data type not inexact
            psi_eps = 0
        try:
            phi_eps = np.finfo(phi.dtype).eps
        except ValueError:
            # data type not inexact
            phi_eps = 0

        eps = max(psi_eps, phi_eps) * np.prod(psi.shape) * phi.shape[-1] * 2

        if not normalized:
            # normalization introduces more floating point error
            eps *= (np.prod(psi.shape) * phi.shape[-1] * 2) ** 2

    if psi.ndim - psi.shape.count(1) == 1:
        # Vector
        inner_product = (psi.T.conj() @ phi).squeeze()
        if normalized:
            norm = 1
        else:
            norm = (linalg.norm(psi) * linalg.norm(phi)).squeeze()
    elif psi.ndim == 2:
        inner_product = dot_HS(psi, phi, eps)
        # Matrix
        if normalized:
            norm = 1
        else:
            norm = np.sqrt(dot_HS(psi, psi, eps) * dot_HS(phi, phi, eps))
    else:
        raise ValueError('Invalid dimension')

    phase = np.angle(inner_product)
    modulus = abs(inner_product)

    return abs(norm - modulus) <= eps, phase


def dot_HS(U: ndarray,
           V: ndarray,
           eps: float = None) -> float:
    r"""Return the Hilbert-Schmidt inner product of U and V,

    .. math::
        \langle U, V\rangle_\mathrm{HS} := \mathrm{tr}(U^\dagger V).

    Parameters
    ----------
    U, V : ndarray
        Objects to compute the inner product of.

    Returns
    -------
    result : float, complex
        The result rounded to precision eps.

    Examples
    --------
    >>> import qutil.qi
    >>> U, V = qutil.qi.paulis[1:3]
    >>> print(dot_HS(U, V))
    0.0
    >>> print(dot_HS(U, U))
    2
    """
    if eps is None:
        # Tolerance is the dtype precision times the number of flops for the
        # matrix multiplication times two to be on the safe side
        try:
            eps = np.finfo(U.dtype).eps * np.prod(U.shape) * V.shape[-1] * 2
        except ValueError:
            # dtype is int and therefore exact
            eps = 0

    if eps == 0:
        decimals = 0
    else:
        decimals = abs(int(np.log10(eps)))

    res = np.round(np.einsum('...ij,...ij', U.conj(), V), decimals)
    return res if res.imag.any() else res.real


def sparsity(arr: ndarray, eps: float = None) -> float:
    """
    Return the sparsity of the array *arr*.

    Parameters
    ----------
    arr: array_like
        The array
    eps: float (default: dtype eps)
        Entries smaller than this value will be treated as zero

    Returns
    -------
    sparsity: float
        The sparsity of the array
    """
    eps = np.finfo(arr.dtype).eps if eps is None else eps
    return (np.abs(arr) <= eps).sum() / arr.size


def density(arr: ndarray, eps: float = None) -> float:
    """
    Return the density of the array *arr*.

    Parameters
    ----------
    arr: array_like
        The array
    eps: float (default: dtype eps)
        Entries smaller than this value will be treated as zero

    Returns
    -------
    density: float
        The density of the array
    """
    eps = np.finfo(arr.dtype).eps if eps is None else eps
    return (np.abs(arr) > eps).sum() / arr.size



from .core import GateLayout

try:
    import qcodes as qc
except ImportError as ie:
    raise ImportError('This module requires qcodes.') from ie


class QcodesGateLayout(GateLayout):
    def __init__(self, gate_to_parameter_mapping, layout_file=None, gate_names=None,
                 gate_mask=None, background_color='#ffffff', foreground_color='tab:gray',
                 cmap='hot', v_min=-2, v_max=0, explode_factor=0.1, fignum=998, text_kwargs=None,
                 offset=(0., 0.)):
        super().__init__(layout_file, gate_names, gate_mask, background_color, foreground_color,
                         cmap, v_min, v_max, explode_factor, fignum, text_kwargs, offset)

        class MaskedGate(qc.Parameter):
            def __init__(self, name, *args, **kwargs):
                super().__init__(name, *args, **kwargs)
            def get_raw(self):
                return 0

        self.gate_to_parameter_mapping = {
            gate: gate_to_parameter_mapping.get(gate, MaskedGate(gate))
            for gate in self.gate_names
        }
        self.gate_parameters = self.gate_to_parameter_mapping.values()

    def get_voltages(self, force: bool = False):
        for i, param in enumerate(self.gate_parameters):
            if force:
                self._latest_voltages[i] = param.get()
            else:
                self._latest_voltages[i] = param.cache.get()



"""This module contains some useful plotting functions.

Moreoever, there are custom stylefiles defined in the ``stylelib``
subdir. You can employ these globally by calling::

    plt.style.use('qutil.plotting.<stylename>')

or within a context::

    with plt.style.context('qutil.plotting.<stylename>'):
        ...

Note that this requires ``matplotlib>=3.7.0``. If you don't want to
upgrade, you can also copy the style files to
``matplotlib.get_configdir()/stylelib``.

Examples
--------
Note that this syntax requires ``matplotlib>=3.7.0``.

Plot data using a style adjusted to APS journals:

>>> import matplotlib.pyplot as plt
>>> with plt.style.context('qutil.plotting.publication_aps_tex'):
...    plt.plot([1, 2], [3, 4], label='presentation style')
...    plt.legend()
...    plt.show(block=False)
[...

Plot data using all available custom styles:

>>> import pathlib, qutil
>>> module_path = pathlib.Path(qutil.__file__).parent
>>> for file in (module_path / 'plotting').glob('*.mplstyle'):
...     file = file.relative_to(module_path)
...     style = '.'.join(('qutil',) + file.parent.parts + (file.stem,))
...     with plt.style.context(style):
...         plt.plot([1, 2], [3, 4])
...         plt.title(file.stem)
...         plt.show(block=False)
[...
"""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach_stub(__name__, __file__)

del lazy



"""This module contains utility functions that help interacting with matlab and matlab files"""
import dataclasses
import itertools
import warnings
import os
import json
import re
import datetime
import pathlib
import array
from typing import Tuple, Sequence, Union, Any, Optional, List, Literal

import hdf5storage

import numpy as np
import pandas as pd
import xarray as xr

import qutil.caching

SM_DATE_FORMAT = '%Y_%m_%d_%H_%M_%S'
SM_DATE_REGEX = re.compile(r'(\d\d\d\d(?:_\d\d){5})')

try:
    import matlab.engine

    try:
        # legacy version shipped with old MATLAB installations
        from matlab import mlarray
    except ImportError:
        # pypi version
        mlarray = None

except (ImportError, OSError):
    warnings.warn("Matlab engine  interface not installed. "
                  "Some functionality requires using MATLAB directly.\n"
                  "If you are using the newest MATLAB you can install it directly from pypi via "
                  "'python -m pip install matlabengine'.\n"
                  r"For older MATLAB versions, navigate to "
                  r"'C:\Program Files\MATLAB\$YOUR_VERSION$\extern\engines\python' and execute "
                  r"'python setup.py install'.",
                  category=ImportWarning)
    matlab = None
    mlarray = None

__all_ = ['load_special_measure_scan', 'cached_load_mat_file', 'special_measure_to_dataframe',
          'load_special_measure_with_matlab_engine', 'mlarray_to_numpy']


if mlarray is not None:
    DTYPE_TO_MLARRAY = {
        np.dtype('float64'): mlarray.double
    }
elif matlab is not None and mlarray is None:  # matlab.mlarray deprecated for versions > 2021a
    DTYPE_TO_MLARRAY = {
        np.dtype('float64'): matlab.double
    }

def mlarray_to_numpy(d) -> np.ndarray:
    """Converts a mlarray (matlab engine array) into a readable numpy array in an efficient manner if possible.

    Calling np.asarray(your_matlab_array) is very slow because mlarray does not implement the buffer protocol.
    This function utilizes the internal private representation if possible
    """
    try:
        data = d._data
        strides = d._strides
        shape = d._size
        item_size = data.itemsize

        return np.lib.stride_tricks.as_strided(
            data,
            shape=shape, strides=[s * item_size for s in strides], writeable=False
        )

    except AttributeError:
        # slow fallback
        return np.asarray(d)


def numpy_to_mlarray(a: np.ndarray):
    mf_type = DTYPE_TO_MLARRAY[a.dtype]
    try:
        d = mf_type(size=a.shape)
        d._data = array.array(d._data.typecode, a.tobytes(order='F'))
        return d
    except AttributeError:
        # fallback
        return mf_type(initializer=a.tolist())


class ModuleEngineWrapper:
    """The purpose of this class is to be a default argument for engine requiring functions different from None.
     This is the least stupid default interface I came up with (Simon)."""
    ENGINE = None

    @classmethod
    def get_engine(cls):
        if cls.ENGINE is None:
            cls.ENGINE = matlab.engine.connect_matlab()
        return cls.ENGINE

    @staticmethod
    def to_engine(obj) -> 'matlab.engine.MatlabEngine':
        if isinstance(obj, str):
            return matlab.engine.connect_matlab(name=obj)
        elif obj is None:
            print('Connecting to (and maybe starting) MATLAB')
            return matlab.engine.connect_matlab()
        elif isinstance(obj, matlab.engine.MatlabEngine):
            return obj
        else:
            return obj.get_engine()


def read_table(engine, table_path: str) -> pd.DataFrame:
    """Read a table from the given path in the engine namespace and return it
    as a pd.DataFrame."""
    row_names = engine.eval(f'{table_path}.Properties.RowNames')
    col_names = engine.eval(f'{table_path}.Properties.VariableNames')
    values = engine.eval(f'{table_path}.Variables')
    return pd.DataFrame(np.array(values),
                            columns=col_names, index=row_names)


def _has_trafofn(engine, scan_path: str) -> bool:
    return engine.eval(f"isfield({scan_path}.loops, 'trafofn') && "
                       f"any(arrayfun(@(loop) ~isempty(loop.trafofn), {scan_path}.loops))")


def _get_config_channel_values(engine, loaded_path: Optional[str]) -> pd.Series:
    if loaded_path:
        configch = engine.eval(f'{loaded_path}.configch')
        configvals = engine.eval(f'{loaded_path}.configvals')
    else:
        configch = engine.workspace['configch']
        configvals = engine.workspace['configvals']
    return pd.Series(np.array(configvals).ravel(), index=configch)



def _mat_string_to_str(mstring) -> str:
    if isinstance(mstring, str):
        return mstring
    elif isinstance(mstring, list):
        return '\n'.join(str(line) for line in mstring)
    else:
        return str(mstring) if len(mstring) else ''


@dataclasses.dataclass
class PlottedData:
    """Representation of the actually plotted object in a matlab figure. Currently image, text and line are supported.
    """

    plot_type: Literal['image', 'text', 'line']

    x_data: Optional[np.ndarray] = dataclasses.field(default=None)
    y_data: Optional[np.ndarray] = dataclasses.field(default=None)
    c_data: Optional[np.ndarray] = dataclasses.field(default=None)

    line_style: Optional[str] = dataclasses.field(default=None)
    text: Optional[Sequence[str]] = dataclasses.field(default=None)

    @classmethod
    def load_from_path(cls, engine, path: str):
        plot_type = engine.eval(f"{path}.Type", nargout=1).lower()

        kwargs = {}
        if engine.eval(f"isfield({path}, 'CData')"):
            kwargs['c_data'] = mlarray_to_numpy(engine.eval(f"{path}.CData", nargout=1)).astype(float)
        if engine.eval(f"isfield({path}, 'XData')"):
            kwargs['x_data'] = mlarray_to_numpy(engine.eval(f"{path}.CData", nargout=1)).astype(float)
        if engine.eval(f"isfield({path}, 'YData')"):
            kwargs['y_data'] = mlarray_to_numpy(engine.eval(f"{path}.CData", nargout=1)).astype(float)

        if engine.eval(f"isfield({path}, 'String')"):
            kwargs['text'] = _mat_string_to_str(engine.eval(f"{path}.String", nargout=1))

        if engine.eval(f"isfield({path}, 'LineStyle')"):
            kwargs['line_style'] = engine.eval(f"{path}.LineStyle", nargout=1)

        if plot_type not in typing.get_args(cls.plot_type):
            warnings.warn(f"Don't know what {plot_type} is.")
        return cls(plot_type, **kwargs)


@dataclasses.dataclass
class AxesData:
    """Data of an axes object in a matlab plot."""
    title: str
    subtitle: str
    xlim: Tuple[float, float]
    ylim: Tuple[float, float]
    xlabel: str
    ylabel: str
    xscale: str
    yscale: str

    content: List[PlottedData]

    @classmethod
    def load_from_path(cls, engine, path: str):
        title = _mat_string_to_str(engine.eval(f"{path}.Title.String", nargout=1))
        if engine.eval(f"{path}.Subtitle.String", nargout=1):
            subtitle = _mat_string_to_str(engine.eval(f"{path}.Subtitle.String", nargout=1))
        else:
            subtitle = None
        xlim = tuple(mlarray_to_numpy(engine.eval(f"{path}.XLim", nargout=1)).astype(float).reshape((-1)))
        ylim = tuple(mlarray_to_numpy(engine.eval(f"{path}.YLim", nargout=1)).astype(float).reshape((-1)))
        xlabel = _mat_string_to_str(engine.eval(f"{path}.XLabel.String", nargout=1))
        ylabel = _mat_string_to_str(engine.eval(f"{path}.YLabel.String", nargout=1))
        xscale = _mat_string_to_str(engine.eval(f"{path}.XScale", nargout=1))
        yscale = _mat_string_to_str(engine.eval(f"{path}.YScale", nargout=1))

        n_children = int(engine.eval(f"length({path}).Children)", nargout=1))
        content = []
        for j in range(n_children):
            child_path = f"{path}.Children({j + 1})"
            child_data = PlottedData.load_from_path(engine, child_path)
            content.append(child_data)

        return cls(title=title, subtitle=subtitle,
                   xlim=xlim, ylim=ylim, xlabel=xlabel, ylabel=ylabel,
                   xscale=xscale, yscale=yscale, content=content)


@dataclasses.dataclass
class FigureData:
    """Data of a matlab figure."""
    name: str
    axes_data: Optional[List[AxesData]]
    user_data: Optional[Any]

    @classmethod
    def load_from_file(cls, file_name: str, engine = ModuleEngineWrapper,
                       fig_path: str = 'loaded_figure',
                       load_axes_data: bool = True,
                       user_data_loader = 'json',
                       user_data_blacklist: Sequence[str] = ()):
        engine = ModuleEngineWrapper.to_engine(engine)


        engine.eval(f"{fig_path} = openfig('{file_name}', 'invisible');", nargout=0)
        name = _mat_string_to_str(f"{fig_path}.Name")
        axes_data = None
        if load_axes_data:
            axes_data = []
            n_axes = n_plots = int(engine.eval(f"length({fig_path}.Children)", nargout=1))
            for ax_idx in range(n_axes):
                ax_data = AxesData.load_from_path(engine, f"{fig_path}.Children({ax_idx + 1})")
                axes_data.append(ax_data)

        user_data = None
        if not user_data_loader:
            pass
        if user_data_loader == 'json':
            user_data = {}
            user_data_fields = list(str(s) for s in engine.eval(f"fieldnames({fig_path}.UserData)'"))
            for fieldname in user_data_fields:
                if fieldname in user_data_blacklist:
                    continue
                try:
                    json_encoded = engine.eval(f'jsonencode({fig_path}.UserData.{fieldname})', nargout=1)
                except Exception as err:
                    warnings.warn(f"Ignoring field '{fieldname}': '{err}'", RuntimeWarning)
                else:
                    value = json.loads(json_encoded)
                    user_data[fieldname] = value
        return cls(name, axes_data=axes_data, user_data=user_data)



def load_data_from_matlab_figure(file_name: os.PathLike,
                                 engine=ModuleEngineWrapper,
                                 with_scan=True):
    """
    Loads data saved in matlab figures

    Args:
        file_name: fig file to load
        engine:
            None, str -> passed to matlab.engine.connect_matlab()
            () -> Use module instance
            MatlabEngine -> used directly

    Returns:
        fig_data
        user_data

    """
    engine = ModuleEngineWrapper.to_engine(engine)

    engine.eval(f"fig = openfig('{file_name}', 'invisible');", nargout=0)
    engine.eval(f"plot = fig.Children;", nargout=0)
    n_plots = int(engine.eval("length(plot)", nargout=1))

    fig_data = []

    for i in range(n_plots):
        n_children = int(engine.eval(f"length(plot({i+1}).Children)", nargout=1))
        if n_children > 0:
            title = engine.eval(f"plot({i+1}).Title.String", nargout=1)
            if isinstance(title, list):
                title = [str(e) for e in title if len(e)]
            else:
                title = str(title) if len(title) else ''

            plot_data = {
                'title': title,
                'subtitle': engine.eval(f"plot({i+1}).Subtitle.String", nargout=1),
                'XLim': mlarray_to_numpy(engine.eval(f"plot({i+1}).XLim", nargout=1)).astype(float).reshape((-1)),
                'YLim': mlarray_to_numpy(engine.eval(f"plot({i+1}).YLim", nargout=1)).astype(float).reshape((-1)),
                'XLabel': str(engine.eval(f"plot({i+1}).XLabel.String", nargout=1)),
                'YLabel': str(engine.eval(f"plot({i+1}).YLabel.String", nargout=1)),
                'XScale': str(engine.eval(f"plot({i+1}).XScale")),
                'YScale': str(engine.eval(f"plot({i+1}).YScale")),
            }
            """
            other properties to potentially use:

            XTick
            YTick
            XAxis
            YAxis
            """

            plot_data['content'] = []
            for j in range(n_children):
                child = {
                    'type': engine.eval(f"plot({i+1}).Children({j+1}).Type", nargout=1).lower()
                }

                if child['type'] == 'image':
                    child['CData'] = mlarray_to_numpy(engine.eval(f"plot({i+1}).Children({j+1}).CData", nargout=1)).astype(float)
                    child['XData'] = mlarray_to_numpy(engine.eval(f"plot({i+1}).Children({j+1}).XData", nargout=1)).astype(float)
                    child['YData'] = mlarray_to_numpy(engine.eval(f"plot({i+1}).Children({j+1}).YData", nargout=1)).astype(float)
                elif child['type'] == 'text':
                    child['text'] = engine.eval(f"plot({i+1}).Children({j+1}).String", nargout=1)
                    child['text'] = [str(e) for e in child['text'] if len(e)]
                elif child['type'] == 'line':
                    child['XData'] = mlarray_to_numpy(engine.eval(f"plot({i+1}).Children({j+1}).XData", nargout=1)).astype(float)
                    child['YData'] = mlarray_to_numpy(engine.eval(f"plot({i+1}).Children({j+1}).YData", nargout=1)).astype(float)
                    child['LineStyle'] = engine.eval(f"plot({i+1}).Children({j+1}).LineStyle", nargout=1)
                else:
                    warnings.warn(f"Don't know what {child['type']} is. Thus nothing extracted.")

                plot_data['content'].append(child)
            fig_data.append(plot_data)

    fig_data = list(reversed(fig_data))

    engine.eval(f"ud = fig.UserData;", nargout=0)
    if engine.eval("isfield(ud, 'scan')") and not with_scan:
        engine.eval(f"ud = rmfield(ud, 'scan');", nargout=0)
    engine.eval(f"ud = jsonencode(ud);", nargout=0)
    user_data = json.loads(engine.workspace['ud'])

    return fig_data, user_data


def _parse_orthogonal_qupulse_scan(scan_path: str, engine: 'matlab.engine.MatlabEngine') -> Tuple[list, list, list, dict]:
    """This function does not really reflect a robust or correct way to extract the information, but it just works most
    of the time."""
    awg_program = json.loads(engine.eval(f"jsonencode({scan_path}.data.awg_program)"))

    # scan_name = awg_program["pulse_template"]["main"]
    pulse_is_named = True

    # maybe one could take the information which pulse template to use from awg_program["pulse_template"]["main"].
    scan_name = engine.eval(f'{scan_path}.data.conf_seq_args.pulse_template', nargout=1)
    if engine.isempty(scan_name):
        # if scan.data.conf_seq_args.pulse_template is empty, one will use something else:
        if 'main' in awg_program["pulse_template"]:
            scan_name = awg_program["pulse_template"]["main"]

    # ok, then the scan does not have a name and we hope that awg_program["parameters_and_dicts"]
    # still contains the necessary information
    if len(scan_name) == 0:
        scan_name = ""
        pulse_is_named = False

    # there are multiple dicts given for the pulse parameters.
    # The later ones overwrite the settings of the prior ones. This is done in the next lines of code.
    scan_params = {}
    for d in awg_program["parameters_and_dicts"]:
        if pulse_is_named and scan_name in d:
            scan_params = {**scan_params, **d[scan_name]}
        else:
            scan_params = {**scan_params, **d}
        
    # the hacky way is now to look for start_x, ..., stop_y within the scan_params
    keywords_of_interest = ['start', 'stop', 'N']
    found_kws = {}
    for kw in keywords_of_interest:
        for sp in scan_params.keys():
            if kw in sp:
                # the keyword of interest is in the selected scan_param
                se = sp.replace(scan_name, "").replace(kw, "").split("_")
                axis_name = "_".join([x for x in se if (x != "")])

                # save that to the dict
                found_kws.setdefault(kw, {})
                found_kws[kw].setdefault(axis_name, [])
                found_kws[kw][axis_name].append(sp)

    prefix_priority = [scan_name, ""]
    selected_kws = {}
    for kw, v in found_kws.items():
        selected_kws[kw] = {}
        for ax, a in v.items():
            a_sorted = [*a]
            a_sorted.sort(key=len)
            for pp in prefix_priority:
                for e in a_sorted:
                    if pp in e:
                        # now the interesting keywords are found, the corresponding values is to be obtained
                        # selected_kws[kw][pf] = v2[pp][-1] # set the keyword
                        selected_kws[kw][ax] = scan_params[e] # set the value
                        break
                if ax in selected_kws[kw]:
                    break
            else:
                warnings.warn(f"The parameter for the keyword {kw} and the axis {ax} has not been found to match with prefix_priority. Will use the shortest one instead.")
                selected_kws[kw][ax] = scan_params[a_sorted[0]]

    # checking if every axis has all the necessary entries
    all_axes = {}
    for k, v in selected_kws.items():
        for kk in v.keys():
            all_axes.setdefault(kk, 0)
            all_axes[kk] += 1
    _v = -1
    for v in all_axes.values():
        if _v == -1:
            _v = v
        elif _v != v:
            warnings.warn(f"incomplete information about scan axes.")

    # TODO need to throw out the entries that do not contain all of the keywords_of_interest values
    # TODO this might need to use user specific parameters.

    # filling the output arrays (could also be more dynamic)
    qupulse_rngs = np.full((len(list(all_axes.keys())), 2), np.nan).astype(float)
    qupulse_npoints_list = np.full(len(list(all_axes.keys())), np.nan).astype(int)
    qupulse_setchans = list(all_axes.keys())
    for i, c in enumerate(qupulse_setchans):
        qupulse_rngs[i][0] = selected_kws['start'][c]
        qupulse_rngs[i][1] = selected_kws['stop'][c]
        qupulse_npoints_list[i] = selected_kws['N'][c]

    return list(qupulse_rngs), list(qupulse_npoints_list), list(qupulse_setchans), dict(
        channel_mapping=awg_program['channel_mapping'],
        global_transformation=awg_program['global_transformation']
    )


def load_special_measure_with_matlab_engine(file_name: os.PathLike,
                                            engine=ModuleEngineWrapper,
                                            return_disp: bool=False,
                                            return_rng_as_linspace_params: bool=False) -> Union[
    Tuple[pd.DataFrame, Sequence[np.ndarray], pd.Series, Sequence[Sequence[str]]],
    Tuple[pd.DataFrame, Sequence[np.ndarray], pd.Series, Sequence[Sequence[str]], dict],
]:
    """
    Load special measure scan using MATLAB. This requires that the package delivered with MATLAB is installed.

    Args:
        file_name: mat file to load
        engine:
            None, str -> passed to matlab.engine.connect_matlab()
            () -> Use module instance
            MatlabEngine -> used directly
        return_disp: Additionally return the disp field of the scan
        return_rng_as_linspace_params: Return the scan axes as linspace parameters (breaks virtual scans)

    Returns:
        scan_axes
        data
        config
        getchans
    """
    if matlab is None:
        raise RuntimeError("Requires MATLAB engine interface.")

    engine = ModuleEngineWrapper.to_engine(engine)

    def normalize_chan(chan) -> Sequence[str]:
        if isinstance(chan, str):
            return [chan]
        elif len(chan) == 0:
            # {} or []
            return []
        else:
            return chan

    # we cannot return a struct array to python so we load it into the namespace
    engine.load(os.fspath(file_name), 'scan', 'data', 'configch', 'configvals', nargout=0)

    for f in ['scan', 'data', 'configch', 'configvals']:
        if not engine.eval(f"exist('{f}', 'var')"):
            raise ValueError(f"{file_name} does not contain {f}.")

    data = engine.workspace['data']
    configch = engine.workspace['configch']
    configvals = engine.workspace['configvals']

    config = pd.Series(np.array(configvals).ravel(), index=configch)

    n_loops = int(engine.eval('numel(scan.loops)'))
    getchans = []
    for ii in range(n_loops):
        getchans.append(normalize_chan(engine.eval(f'scan.loops({ii + 1}).getchan')))

    rngs = []
    npoints_list = []
    setchans = []

    for ii in range(n_loops):
        rng = np.array(engine.eval(f'scan.loops({ii+1}).rng')).ravel()
        if engine.eval(f"isfield(scan.loops({ii + 1}), 'npoints') && ~isempty(scan.loops({ii + 1}).npoints)"):
            npoints = int(engine.eval(f'scan.loops({ii + 1}).npoints'))
        else:
            npoints = rng.size

        if not len(rng) == 0:
            # TODO rng==[] could be used as a flag to look for awg pulses. 
            rngs.append(rng)
            npoints_list.append(npoints)
            setchans.append(normalize_chan(engine.eval(f'scan.loops({ii + 1}).setchan')))

    # hacked together method for detecting qupulse pulses:
    # looking for awg program
    has_pulse_info = engine.eval("isfield(scan, 'data') && isfield(scan.data, 'awg_program')", nargout=1)
    if has_pulse_info:
        qupulse_rngs, qupulse_npoints_list, qupulse_setchans, attrs = _parse_orthogonal_qupulse_scan('scan', engine)
        rngs = qupulse_rngs + rngs
        npoints_list = qupulse_npoints_list + npoints_list
        setchans = qupulse_setchans + setchans
        config.attrs.update(attrs)

    # process trafofn
    if engine.eval("isfield(scan.loops, 'trafofn') && any(arrayfun(@(loop) ~isempty(loop.trafofn), scan.loops))"):
        try:
            n_setchans = int(engine.eval('numel(scan.loops(1).trafofn)'))
            _scan_axes = []
            for ii in range(n_setchans):
                _scan_axes.append(read_table(engine, f'scan.loops(1).trafofn({ii+1}).args{{1}}'))
            scan_axes = pd.concat(_scan_axes, axis='columns')

        except matlab.engine.MatlabExecutionError as e:
            warnings.warn(f"The used trafofns are not understood. And applying them to the extracted ranges and stuff is not implemented.")
            warnings.warn(f"matlab.engine.MatlabExecutionError: {str(e)}")

    if not return_rng_as_linspace_params:
        scan_axes_col = list({chan for chans in setchans for chan in chans})
        scan_axes_rows = [('origin', 0)]
        for l, rng in enumerate(rngs):
            scan_axes_rows.extend((f'loop_{l + 1}', jj) for jj in range(1, rng.size))
        scan_axes_rows = pd.MultiIndex.from_tuples(scan_axes_rows, names=('axis', 'n'))

        scan_axes = pd.DataFrame(index=scan_axes_rows, columns=scan_axes_col)

        for col in scan_axes.columns:
            for idx, setchan in enumerate(setchans):
                if col in setchan:
                    axis = f'loop_{idx+1}'
                    for n, x in enumerate(rngs[idx]):
                        if n == 0:
                            scan_axes.loc[('origin', 0), col] = x
                        else:
                            scan_axes.loc[(axis, n), col] = x
                    break
    else:
        scan_axes = {}
        for i, c in enumerate(setchans):
            if isinstance(c, list):
                _c = ",".join(c)
            else:
                _c = c
            scan_axes[_c] = (*rngs[i], npoints_list[i])

    if return_disp:
        disp = json.loads(engine.eval("jsonencode(scan.disp)", nargout=1))
        return scan_axes, [mlarray_to_numpy(d) for d in data], config, getchans, disp
    else:
        return scan_axes, [mlarray_to_numpy(d) for d in data], config, getchans


def special_measure_to_dataframe(loaded_scan_data: dict,
                                 squeeze_constant_setchan: bool = True) -> pd.DataFrame:
    """Try to interpret the data returned from hdf5storage.loadmat(filename)
    as a pd.DataFrame.

    Not handled/tested yet:
        - Buffered measurements
        - trafofn
        - procfn
        - any thing with MATLAB tables or other classes.
    """
    scan = loaded_scan_data['scan']
    assert scan.shape == (1, 1)
    scan = scan[0, 0]

    loops = scan['loops']
    assert len(loops.shape) == 2
    assert loops.shape[0] == 1
    loops = loops[0, :]

    n_loops = loops.size

    # fails if a loop has more than one npoints
    npoints = list(loops['npoints'].astype(np.int64))
    for idx, npoint in enumerate(npoints):
        if npoint < 0:
            warnings.warn(f"Negative npoints {npoint} in loop {idx} gets clamped to 0")
            npoints[idx] = max(npoint, 0)

    setchan = list(loops['setchan'])
    for loop_idx, chan in enumerate(setchan):
        assert len(chan.shape) == 2
        assert chan.shape[0] == 1
        setchan[loop_idx] = tuple(ch[0] for ch in chan.ravel())

    # rngs can be per channel or
    rngs = list(loops['rng'])
    for loop_idx, rng in enumerate(rngs):
        assert rng.shape == (1, 2)
        rngs[loop_idx] = tuple(rng.ravel())

    # This code needs to be adapted if it is possible to use different ranges
    # for different setchannels in the same loop. This might require
    # interpreting the trafofn
    sweeps = []
    for loop_idx in range(n_loops):
        loop_sweeps = {}

        span = np.linspace(*rngs[loop_idx], num=npoints[loop_idx])

        for ch in setchan[loop_idx]:
            loop_sweeps[ch] = span
        sweeps.append(loop_sweeps)

    labels, values = [], []
    for sweep in sweeps:
        if len(sweep) > 1:
            vals = list(sweep.values())
            if np.unique(vals, axis=0).shape[0] != 1:
                raise RuntimeError('Simultaneous sweep with different ranges not supported', vals)
            labels.append('-'.join(sweep.keys()))
            values.append(vals[0])

    idx = pd.MultiIndex.from_product(values, names=labels)

    # buffered measurements no handled?
    getchan = loops['getchan']
    for loop_idx, chan in enumerate(getchan):
        assert len(chan.shape) == 2
        assert chan.shape[0] == 1
        getchan[loop_idx] = tuple(ch[0] for ch in chan.ravel())

    measured = list(itertools.chain.from_iterable(getchan))

    # always vector cell
    data = loaded_scan_data['data']
    assert data.shape == (1, len(measured))
    data = data[0, :]

    result = pd.DataFrame(index=idx)
    assert len(measured) == len(data)
    for meas, val in zip(measured, data):
        val = val.transpose()
        result[meas] = val.flatten()
        assert result[meas].shape == idx.levshape

    if squeeze_constant_setchan:
        to_squeeze = [lvl_idx
                      for lvl_idx, lvl_dim in enumerate(idx.levshape)
                      if lvl_dim == 1]
        result = result.droplevel(to_squeeze)

    try:
        result.attrs['consts'] = loaded_scan_data['scan']['consts']
    except ValueError:
        pass

    return result


@dataclasses.dataclass
class SpecialMeasureScan:
    """Baseline data of special measure scans.

    TODO: store and interpret the disp attribute

    Attributes:
        config: special measure channels and values at the time of scan start.
        data: Measured data field in mat file
        save_time: Timestamp extracted from file metadata i.e. the modification time
        meas_time: Timestamp extracted from the file name if possible
        name: File name without the extension
    """

    config: pd.Series
    data: Tuple[np.ndarray, ...]
    save_time: Optional[pd.Timestamp]
    meas_time: Optional[pd.Timestamp]
    name: str

    @property
    def attrs(self):
        attrs = dict(config=self.config)
        if self.meas_time is not None:
            attrs['meas_time'] = self.meas_time
        if self.save_time is not None:
            attrs['save_time'] = self.save_time
        return attrs


@dataclasses.dataclass
class SimpleScan(SpecialMeasureScan):
    """Represents a simple special measure scan.

    Assumptions:
         - Each loop has one or more setchannels
         - No trafofn
         - Procfn results are ignored but may be included
         - All get channels are in the fast loop
    """

    # per loop one setchannel. Fast axis is setchans[-1]
    set_channels: Sequence[str]
    set_ranges: Sequence[Sequence[float]]

    # only fast loop has getchannels
    get_channels: Tuple[str, ...]
    
    @property
    def main_channels(self) -> Tuple[str, ...]:
        return tuple(self.set_channels)
        

    def to_xarray(self, procfn='warn') -> xr.DataArray:
        assert procfn in ('ignore', 'warn', 'error')
        assert len(self.get_channels) <= len(self.data)
        if len(self.get_channels) < len(self.data):
            if procfn == 'warn':
                warnings.warn("Scan apparently has procfn results")
            elif procfn == 'error':
                raise ValueError("Scan apparently has procfn results")
            data = self.data[:len(self.get_channels)]
        else:
            data = self.data
        shape = tuple(len(ax) for ax in self.set_ranges)
        for d in data:
            assert d.shape == shape

        # fast axis is
        coords = [('getchan', list(self.get_channels))]
        coords.extend(
            (setchan, np.asarray(values))
            for setchan, values in zip(self.set_channels, self.set_ranges)
        )

        return xr.DataArray(
            np.stack(data, axis=0),
            coords=coords,
            attrs=self.attrs,
            name=self.name
        )


@dataclasses.dataclass
class VirtualScan(SpecialMeasureScan):
    """Virtual i.e. non-orthogonal scan.

    The virtual scan is in the space of `physical_gates`. It starts at `origin` and the axes that span the scan are
    in `scan_axes`. The fastest is `scan_axes[-1]`. Currently only data that corresponds to all `get_channels` being
    measured at all points is allowed."""
    physical_gates: Tuple[str, ...]
    origin: Tuple[float, ...]
    scan_axes: Sequence[Tuple[float, ...]]

    # only fast loop has getchannels
    get_channels: Tuple[str, ...]

    @property
    def main_channels(self) -> Tuple[str, ...]:
        """Return the names of the channels which have the highest magnitude in each axis."""
        return tuple(self.physical_gates[max(range(len(axis)), key=lambda idx: abs(axis[idx]))]
                     for axis in self.scan_axes)

    def to_xarray(self):
        assert len(self.get_channels) <= len(self.data)
        if len(self.get_channels) < len(self.data):
            warnings.warn("Scan apparently has procfn results")

        if len(self.scan_axes) == 1 and all(len(d.shape) > 1 and d.shape[0] == 1 for d in self.data):
            data = [d[0, ...] for d in self.data]
        else:
            data = self.data
        
        shape, = set(d.shape for d in data[:len(self.get_channels)])
        if len(shape) != len(self.scan_axes):
            raise ValueError("Malformed virtual scan data")

        coords = [('getchan', list(self.get_channels))]
        for n_points, scan_axis in zip(shape, self.scan_axes):
            max_idx = int(np.argmax(np.abs(scan_axis)))

            axis_values = self.origin[max_idx] + np.linspace(0., scan_axis[max_idx], num=n_points)

            axis_name = []
            for gate, value in zip(self.physical_gates, scan_axis):
                if np.abs(value) > 1e-15:
                    axis_name.append(f'{value}*{gate}')
            axis_name = f'Virtual {self.physical_gates[max_idx]}:' + ','.join(axis_name)

            coords.append((axis_name, axis_values))

        return xr.DataArray(
            np.stack(data, axis=0),
            coords=coords,
            attrs={**self.attrs, 'origin': self.origin},
            name=self.name
        )


def _read_common_fast_loop_getchan_fields(engine, name_for_loaded: str, shape: Tuple[int, ...]):
    data = engine.eval(f'{name_for_loaded}.data')
    data = [mlarray_to_numpy(d) for d in data]
    squeeze_axis = tuple(idx for idx, dim in enumerate(shape) if dim == -1)
    if squeeze_axis:
        for idx, d in enumerate(data):
            try:
                data[idx] = np.squeeze(d, squeeze_axis)
            except ValueError:
                data[idx] = d
        shape = tuple(dim for dim in shape if dim != -1)
    elif np.product(shape) == 1:
        data = [d.reshape(shape) for d in data]
    elif len(shape) == 1:
        for idx, d in enumerate(data):
            if shape[0] in d.shape:
                data[idx] = d.reshape(shape)
        
    
    config = _get_config_channel_values(engine, name_for_loaded)
    n_loops = int(engine.eval(f'numel({name_for_loaded}.scan.loops)'))
    if n_loops == 0:
        raise ValueError("No loops defined")
    getchans = engine.eval(f'[{name_for_loaded}.scan.loops.getchan]')
    
    for d in data:
        if d.shape == ():
            raise ValueError('BUG')

    for (d, ch) in itertools.zip_longest(data, getchans, fillvalue=None):
        if d is None:
            raise TypeError("No data for getchannel. Indicates an invalid or not simple scan", ch)
        elif ch is None:
            warnings.warn("Data found that could not be accociated with a get channel. It is ignored")
        elif d.shape != shape:
            raise TypeError("data shape deviates from expected", d.shape, shape)
    return data, config, getchans


def load_simple_scan_with_matlab_engine(file_name: os.PathLike,
                                        engine=ModuleEngineWrapper,
                                        name_for_loaded: str = 'loaded') -> SimpleScan:
    """Load special measure scan using MATLAB engine. This requires that the package delivered with MATLAB is installed.

    Args:
        file_name: mat file to load
        engine:
            None, str -> passed to matlab.engine.connect_matlab()
            () -> Use module instance
            MatlabEngine -> used directly
        name_for_loaded: the given file is loaded into this variable in the matlab workspace

    Returns:
        simple_scan

    Raises:
        ValueError: if scan is not valid
        TypeError: if scan is not simple
    """
    if matlab is None:
        raise RuntimeError("Requires MATLAB engine interface.")

    engine = ModuleEngineWrapper.to_engine(engine)

    # we cannot return a struct array to python so we load it into the namespace
    engine.eval(f'{name_for_loaded} = load("{os.fspath(file_name)}");', nargout=0)
    name = pathlib.Path(file_name)

    if _has_trafofn(engine, f'{name_for_loaded}.scan'):
        raise TypeError("Cannot extract SimpleScan from scan with trafofn")
    if not engine.eval(f"isfield({name_for_loaded}, 'data')"):
        raise ValueError("No data in scan", file_name)

    scan = f'{name_for_loaded}.scan'
    n_loops = int(engine.eval(f'numel({scan}.loops)'))
    if n_loops == 0:
        raise ValueError("No loops defined")

    rngs = []
    setchans = []
    shape = []

    # we reverse here to be in the same order as data dimensions
    for ii in reversed(range(n_loops)):
        loop = f'{scan}.loops({ii+1})'
        
        setchan = _to_str_list(engine.eval(f'{loop}.setchan'))
        if not setchan:
            shape.append(-1)
            continue
            #raise TypeError(f"Loop {loop} (MATLAB index) does not have a setchan.")
        
        rng = np.array(engine.eval(f'{loop}.rng')).ravel()
        if engine.eval(f"isfield({loop}, 'npoints') && ~isempty({loop}.npoints)"):
            npoints = int(engine.eval(f'{loop}.npoints'))
            if rng.size != npoints:
                rng = np.linspace(*rng, num=npoints)


        setchans.append(','.join(setchan))
        rngs.append(rng)
        shape.append(len(rng))

    shape = tuple(shape)
    data, config, getchans = _read_common_fast_loop_getchan_fields(engine, name_for_loaded, shape)

    try:
        meas_time = _get_datetime_from_file_name(name)
    except ValueError:
        meas_time = None

    return SimpleScan(
        name=name.stem,
        data=tuple(data),
        set_ranges=rngs,
        set_channels=setchans,
        get_channels=getchans,
        config=config,
        save_time=pd.Timestamp.fromtimestamp(os.path.getmtime(file_name)),
        meas_time=meas_time
    )


def _parse_tdg_trafofn(engine: 'matlab.engine.MatlabEngine', loops: str) -> pd.DataFrame:
    """

    Args:
        engine:
        loops:

    Raises:
        TypeError: no suitable trafofn

    Returns:

    """
    if not engine.eval(f"isfield({loops}, 'trafofn')"):
        raise TypeError("No trafofn defined")

    n_trafofns = mlarray_to_numpy(engine.eval(f"arrayfun(@(loop) numel(loop.trafofn), {loops})")).ravel()
    
    # we allow the lasts loop trafofn to be empty
    if n_trafofns[-1] == 0:
        n_trafofns = n_trafofns[:-1]
    
    if len(set(n_trafofns)) != 1:
        raise TypeError("The number of trafofn is not uniform accross loops", n_trafofns)
    n_setchan = int(n_trafofns[0])

    # FIXME: this does not work if `loops` is a reference i.e. a return value
    n_loops = len(n_trafofns)
    loop_tables = []
    for ii_loop in range(n_loops):
        loop_table = []
        for jj_setchan in range(n_setchan):
            args_ref = f"{loops}({ii_loop + 1}).trafofn({jj_setchan + 1}).args"
            if engine.eval(f"numel({args_ref})") != 1:
                raise TypeError("At least one trafofn has more than one argument")

            table = read_table(engine, args_ref + "{1}")
            loop_table.append(table)

        scan_axes = pd.concat(loop_table, axis='columns')
        loop_tables.append(scan_axes)
    for ii_loop in range(1, n_loops):
        if not loop_tables[0].equals(loop_tables[ii_loop]):
            raise TypeError("Tables are not uniform across loops")
    return loop_tables[0]


def load_virtual_scan_with_matlab_engine(file_name: os.PathLike,
                                         engine=ModuleEngineWrapper,
                                         name_for_loaded: str = 'loaded_virtual') -> VirtualScan:
    if matlab is None:
        raise RuntimeError("Requires MATLAB engine interface.")

    engine = ModuleEngineWrapper.to_engine(engine)

    # we cannot return a struct array to python so we load it into the namespace
    engine.eval(f'{name_for_loaded} = load("{os.fspath(file_name)}");', nargout=0)
    name = pathlib.Path(file_name)

    if not engine.eval(f"isfield({name_for_loaded}, 'data')"):
        raise ValueError("No data in scan", file_name)

    sm_scan_axes = _parse_tdg_trafofn(engine, f'{name_for_loaded}.scan.loops')
    origin = sm_scan_axes.loc['origin']
    scan_axes = list(map(tuple, sm_scan_axes.iloc[1::, :].values))[::-1]
    physical_gates = tuple(sm_scan_axes.columns)

    shape = tuple(mlarray_to_numpy(engine.eval(f"[{name_for_loaded}.scan.loops.npoints]")).ravel()[::-1].astype(int))
    data, config, getchans = _read_common_fast_loop_getchan_fields(engine, name_for_loaded, shape)

    try:
        meas_time = _get_datetime_from_file_name(name)
    except ValueError:
        meas_time = None

    return VirtualScan(
        data=tuple(data),
        config=config,
        get_channels=getchans,
        name=name.stem,
        origin=origin,
        scan_axes=scan_axes,
        physical_gates=physical_gates,
        save_time=pd.Timestamp.fromtimestamp(os.path.getmtime(file_name)),
        meas_time=meas_time,
    )


def load_special_measure_scan(file_name: os.PathLike,
                              squeeze_constant_setchan: bool = True) -> pd.DataFrame:
    """
    :param file_name: Path of the file to load
    :param squeeze_constant_setchan: If true, "set channels" that are constant are not included in the index
    :return: Data frame with a multi-index that corresponds to the "set channels" and columns that correspond to the
    "get channels".
    """
    file_name = os.fspath(file_name)

    # this is slow as the scan stuct is quite complicated and hdf5storage creates a dtype for the whole thing
    file_contents = hdf5storage.loadmat(file_name)

    return special_measure_to_dataframe(file_contents, squeeze_constant_setchan)


@qutil.caching.file_cache
def cached_load_mat_file(filename):
    return hdf5storage.loadmat(filename)


def _get_datetime_from_file_name(fname: Union[pathlib.Path, str]) -> datetime.datetime:
    # TODO: move somewhere else
    stem = getattr(fname, 'stem', fname)
    m = SM_DATE_REGEX.fullmatch(stem)
    if m:
        return datetime.datetime.strptime(m.group(1), SM_DATE_FORMAT)
    else:
        raise ValueError("Could not find datetime in file name:", fname, stem)


def _to_str_list(obj: Union[str, List[str], 'matlab.mlarray', 'matlab.double']) -> List[str]:
    """
    handles these MATLAB expressions:
        []
        {}
        'asd'
        {'asd', 'dfg'}
    """
    if not obj:
        return []
    if isinstance(obj, (list, tuple)):
        return list(map(str, obj))
    else:
        return [str(obj)]




import warnings

try:
    from python_spectrometer import daq as daq
    from python_spectrometer import Spectrometer as Spectrometer
    warnings.warn('The spectrometer module has moved to its own package and is now called '
                  "'python_spectrometer'. Please update your imports accordingly.",
                  DeprecationWarning)
except ImportError:
    url = 'https://git.rwth-aachen.de/qutech/python-spectrometer'
    warnings.warn(f'The spectrometer module has moved to its own package at {url}. You can "'
                  "install it from there or from pypi via 'pip install python-spectrometer'",
                  # UserWarning so that it isn't suppress by most defaults:
                  # https://docs.python.org/3/library/exceptions.html#ImportWarning
                  UserWarning)
    raise

del warnings



"""Import everything from itertools, more_itertools and some custom functions """
from __future__ import annotations

import warnings
from math import inf
from typing import (Iterable, Any, Type, List, Tuple, Iterator, Protocol, TypeVar, TYPE_CHECKING,
                    Sequence, Callable)

from itertools import *
from more_itertools import *

try:
    from typing import ParamSpec, SupportsAbs
except ImportError:
    from typing_extensions import ParamSpec, SupportsAbs

from . import functools as _functools

_T = TypeVar('_T')
_P = ParamSpec('_P')

if TYPE_CHECKING:
    from _typeshed import SupportsAdd, SupportsSub, SupportsAllComparisons

    class SupportsAddSub(SupportsAdd[Any], SupportsSub[Any], Protocol):
        ...

    class SupportsAddSubComparison(SupportsAddSub[Any], SupportsAllComparisons[Any], Protocol):
        ...

    SupportsAddSubT = TypeVar('SupportsAddSubT', bound=SupportsAddSub)
    SupportsAddSubComparisonT = TypeVar('SupportsAddSubComparisonT',
                                        bound=SupportsAddSubComparison)


def _round(x: float, decimals: int | None = None) -> float | int:
    """Overrides builtin round behavior so that decimals==None does not
    round at all."""
    if decimals is None:
        return x
    return round(x, decimals)


def separate_iterator(it: Iterable, sep: Any) -> Iterator:
    """separate_iterator('abcde', ',') --> a , b , c , d , e

    The same as :func:`~more_itertools.intersperse(sep, it, n=1)`. Only here for backwards compability.
    """
    warnings.warn(f"{__name__}.separate_iterator is just a wrapper around intersperse.",
                  DeprecationWarning,
                  stacklevel=2)
    return intersperse(sep, it, n=1)


def flatten_nested(it: Iterable, dtypes: Tuple[Type[Iterable], ...] = (list,)) -> Iterator:
    """Flattens the given arbitrarily nested iterable. By default, only lists are flattened. Use the optional `dtypes`
    argument to flatten other iterables as well.

    Similar to :func:`~more_itertools.collapse` which works with a blacklist instead of a whitelist.

    Parameters
    ----------
    it: :class:`~typing.Iterable` to flatten
    dtypes: Types that get flattened.

    Returns
    -------
    Objects that are not of a type in `dtypes`
    """
    if isinstance(it, dtypes):
        for x in it:
            yield from flatten_nested(x, dtypes)
    else:
        yield it


def argmin(it: Iterable) -> int:
    """Return index of smallest element by iteration order.

    >>> argmin([1, 3, 2, 4, 0])
    4

    Raises
    --------
    :class:`~ValueError` if the iterable is empty.

    Parameters
    ----------
    it: :class:`~typing.Iterable` to search for minimum

    Returns
    -------
    Index of the smallest element if it by iteration order
    """
    it = enumerate(it)
    try:
        min_idx, current = next(it)
    except StopIteration:
        raise ValueError('Argument was empty')
    for idx, elem in it:
        if elem < current:
            current = elem
            min_idx = idx
    return min_idx


def argmax(it: Iterable) -> int:
    """Return index of largest element by iteration order.

    >>> argmax([1, 3, 2, 4, 0])
    3

    Raises
    --------
    :class:`~ValueError` if the iterable is empty.

    Parameters
    ----------
    it: :class:`~typing.Iterable` to search for maximum

    Returns
    -------
    Index of the largest element if it by iteration order
    """
    it = enumerate(it)
    try:
        max_idx, current = next(it)
    except StopIteration:
        raise ValueError('Argument was empty')
    for idx, elem in it:
        if elem > current:
            current = elem
            max_idx = idx
    return max_idx


def next_largest(it: Iterable[SupportsAddSubComparisonT],
                 value: SupportsAddSubT,
                 precision: int | None = None) -> SupportsAddSubT:
    """Return the next-largest element to value from it.

    Parameters
    ----------
    it: Iterable
        The iterable to search.
    value: number-like
        The number to find the next-largest element to.
    precision: int, optional
        The number of decimal places up to which elements are compared.

    Returns
    -------
    The next-largest element to value in it, or the value itself if it
    is empty.

    Examples
    --------
    >>> next_largest([3, 4, 1, -5.0], 2.7)
    3
    >>> next_largest([3, 4, 1, -5.0], 1.01)
    3
    >>> next_largest([3, 4, 1, -5.0], -10)
    -5.0
    >>> next_largest([3, 4, 1, -5.0], 7.123)
    4
    >>> next_largest([1, 2, 10321], inf)
    10321
    >>> next_largest([], 42)
    42
    >>> next_largest([10, 11.0], 10 + 1e-14)
    11.0
    >>> next_largest([10, 11.0], 10 + 1e-14, precision=13)
    10

    """
    empty_sentinel = object()
    pred_false, pred_true = partition(lambda x: _round(x - value, precision) >= 0, it)
    result = min(pred_true, default=empty_sentinel)
    if result is empty_sentinel:
        return max(pred_false, default=value)
    return result


def next_smallest(it: Iterable[SupportsAddSubComparisonT],
                  value: SupportsAddSubT,
                  precision: int | None = None) -> SupportsAddSubT:
    """Return the next-smallest element to value from it.

    Parameters
    ----------
    it: Iterable
        The iterable to search.
    value: number-like
        The number to find the next-largest element to.
    precision: int, optional
        The number of decimal places up to which elements are compared.

    Returns
    -------
    The next-largest element to value in it, or the value itself if it
    is empty.

    Examples
    --------
    >>> next_smallest([3, 4, 1, -5.0], 2.7)
    1
    >>> next_smallest([3, 4, 1, -5.0], 1.01)
    1
    >>> next_smallest([3, 4, 1, -5.0], -10)
    -5.0
    >>> next_smallest([3, 4, 1, -5.0], 7.123)
    4
    >>> next_smallest([1, 2, 10321], inf)
    10321
    >>> next_smallest([], 42)
    42
    >>> next_smallest([10, 11.0], 11 - 1e-14)
    10
    >>> next_smallest([10, 11.0], 11 - 1e-14, precision=13)
    11.0

    """
    empty_sentinel = object()
    pred_false, pred_true = partition(lambda x: _round(x - value, precision) <= 0, it)
    result = max(pred_true, default=empty_sentinel)
    if result is empty_sentinel:
        return min(pred_false, default=value)
    return result


def next_closest(it: Iterable[SupportsAddSubComparisonT],
                 value: SupportsAddSubT,
                 precision: int | None = None) -> SupportsAddSubT:
    """Return the next-closest element to value from it.

    Parameters
    ----------
    it: Iterable
        The iterable to search.
    value: number-like
        The number to find the next-largest element to.
    precision: int, optional
        The number of decimal places up to which elements are compared.

    Returns
    -------
    The next-closest element to value in it, or the value itself if it
    is empty.

    Examples
    --------
    >>> next_closest([3, 4, 1, -5.0], 2.7)
    3
    >>> next_closest([3, 4, 1, -5.0], 1.01)
    1
    >>> next_closest([3, 4, 1, -5.0], -10)
    -5.0
    >>> next_closest([3, 4, 1, -5.0], 7.123)
    4
    >>> next_closest([1, 2, 10321], inf)
    10321
    >>> next_closest([], 42)
    42
    >>> next_closest([10, 11], 10.5 + 1e-14)
    11
    >>> next_closest([10, 11], 10.5 + 1e-14, precision=13)
    10

    """
    if value == -inf:
        return min(it, default=value)
    elif value == inf:
        return max(it, default=value)
    return min(it, key=lambda x: _round(abs(x - value), precision), default=value)


def replace_except(it: Iterable, exceptions: Any, replacement: Any = None) -> Iterator:
    """Replace all items that cause the specified exceptions during
    iteration with replacement.

    >>> list(replace_except(map(lambda d: 5 / d, [2, 1, 0, 5]), ZeroDivisionError, inf))
    [2.5, 5.0, inf, 1.0]

    Use a unique object to be able to identify items that errored:

    >>> sentinel = object()
    >>> filtered = list(replace_except(map(lambda d: 5 / d, [2, 1, 0, 5]), ZeroDivisionError,
    ...                                    sentinel))
    >>> filtered.index(sentinel)
    2

    Be aware that some iterators do not allow further iteration if they threw an exception.

    >>> list(ignore_exceptions((5 / d for d in [2, 1, 0, 5]), ZeroDivisionError))
    [2.5, 5.0]

    Parameters
    ----------
    it : :class:`~typing.Iterable` to iterate over.
    exceptions : Valid argument to an ``except`` clause
    replacement : Any object that is yielded whenever a specified exception
                  occurs.

    See Also
    --------
    :func:`ignore_exceptions` : Ignore instead of replace items that errored.

    """
    it = iter(it)
    while True:
        try:
            yield from it
        except exceptions:
            yield replacement
        else:
            return


def ignore_exceptions(it: Iterable, exceptions: Any) -> Iterator:
    """Ignore all specified exceptions during iteration.

    >>> list(ignore_exceptions(map(lambda d: 5 / d, [2, 1, 0, 5]), ZeroDivisionError))
    [2.5, 5.0, 1.0]

    Be aware that some iterators do not allow further iteration if they threw an exception.

    >>> list(ignore_exceptions((5 / d for d in [2, 1, 0, 5]), ZeroDivisionError))
    [2.5, 5.0]

    Parameters
    ----------
    it: :class:`~typing.Iterable` to iterate over.
    exceptions: Valid argument to an ``except`` clause
    """
    it = iter(it)
    while True:
        try:
            yield from it
        except exceptions:
            continue
        else:
            return


def batched_variable(iterable: Iterable[_T], ns: Sequence[int],
                     strict: bool = False) -> Iterator[Tuple[_T, ...]]:
    """Break *iterable* into consecutive tuples of *n* items for each
    *n* in *ns*.

    Note that *ns* is repeated to match the length of *iterable*.

    If ``strict=True``, an exception is raised if
    ``len(iterable) != sum(ns)``.

    Examples
    --------
    >>> list(batched_variable(range(12), [2, 3, 1]))
    [(0, 1), (2, 3, 4), (5,), (6, 7), (8, 9, 10), (11,)]
    >>> list(map_if(batched_variable([6, 3, -2, 8, 1, 0], [1, 2]),
    ...             lambda x: len(x) == 2, sum, min))
    [6, 1, 8, 1]
    """
    it = iter(iterable)
    return interleave(*(batched(it, n, strict=strict) for n in ns))


def take_and_consume(n: int, consume_before: int, consume_after: int,
                     iterable: Iterable[_T]) -> List[_T]:
    """Take *n* items from *iterable*, consuming *consume_before* and
    *consume_after* taking.

    Examples
    --------
    >>> it = iter(range(7))
    >>> take_and_consume(2, 1, 3, it)
    [1, 2]
    >>> list(it)
    [6]
    """
    consume(iterable, consume_before)
    vals = take(n, iterable)
    consume(iterable, consume_after)
    return vals


def distribute_variable_batches(iterable: Iterable[_T],
                                ns: Sequence[int]) -> List[Iterator[List[_T]]]:
    """Returns an iterator for each *n* in *ns* that yields lists of
    *n* items from *iterable*.

    That is, the iterable is distributed on weighted by *n*.

    Uses :py:func:`~itertools.tee`, so might consume a lot of memory.

    Examples
    --------
    >>> list(map(list, distribute_variable_batches(range(6), [1, 2])))
    [[[0], [3]], [[1, 2], [4, 5]]]
    >>> list(map(list, map(collapse, distribute_variable_batches(range(10), [3, 2]))))
    [[0, 1, 2, 5, 6, 7], [3, 4, 8, 9]]
    """
    return [iter(_functools.partial(take_and_consume, ns[i], sum(ns[:i]), sum(ns[i+1:]), it), [])
            for i, it in zip(range(len(ns)), tee(iter(iterable), len(ns)))]


def zipstarmap(funcs: Sequence[Callable[_P, _T]], iterable: Iterable[_P]) -> List[Iterator[_T]]:
    """Apply *funcs* alternatingly on *args from *iterable*.

    Uses :py:func:`~itertools.tee`, so might consume a lot of memory.

    Examples
    --------
    >>> from operator import add
    >>> # [[-3 + -2, 0 + 1], [abs(-1), abs(2)]]
    >>> list(map(list, zipstarmap([add, abs], batched_variable(range(-3, 3), [2, 1]))))
    [[-5, 1], [1, 2]]
    """
    n = len(funcs)
    return [starmap(func, it) for func, it in zip(funcs, distribute(n, iterable))]


def zipmap(funcs: Sequence[Callable[_P, _T]], iterable: Iterable[_P]) -> List[Iterator[_T]]:
    """Apply *funcs* alternatingly on args from *iterable*.

    Uses :py:func:`~itertools.tee`, so might consume a lot of memory.

    Examples
    --------
    >>> # [[-3 + -2, 2 + 3], [min([-1, 0, 1]), min([4, 5, 6])]]
    >>> list(map(list, zipmap([sum, min], batched_variable(range(-3, 7), [2, 3]))))
    [[-5, 5], [-1, 4]]
    """
    n = len(funcs)
    sliced_its = (islice(it, start, None, n) for it, start in zip(tee(iterable, n), range(n)))
    return [map(func, it) for func, it in zip(funcs, sliced_its)]


def absmin(iterable_or_value: SupportsAbs[_T] | Iterable[SupportsAbs[_T]],
           *others: SupportsAbs[_T], default: _T = object(), key=None) -> _T:
    """The absolute-value-wise minimum. See :func:`minmax` for usage.

    This function's NumPy equivalent is::

        np.abs(iterable).min()

    Examples
    --------
    >>> absmin([-7, 3, 1, -1])
    1
    >>> absmin(-7, 3, 1, -1)
    1
    >>> import numpy as np
    >>> x = np.random.default_rng().normal(20)
    >>> print(absmin(x) == np.abs(x).min())
    True
    """
    iterable_or_value = (iterable_or_value, *others) if others else iterable_or_value
    try:
        return min((abs(x) for x in iterable_or_value), default=default, key=key)
    except TypeError:
        return abs(iterable_or_value)


def absmax(iterable_or_value: SupportsAbs[_T] | Iterable[SupportsAbs[_T]],
           *others: SupportsAbs[_T], default: _T = object(), key=None) -> _T:
    """The absolute-value-wise maximum. See :func:`minmax` for usage.

    This function's NumPy equivalent is::

        np.abs(iterable).max()

    Examples
    --------
    >>> absmax([-7, 3, 1, -1])
    7
    >>> absmax(-7, 3, 1, -1)
    7
    >>> import numpy as np
    >>> x = np.random.default_rng().normal(20)
    >>> print(absmax(x) == np.abs(x).max())
    True
    """
    iterable_or_value = (iterable_or_value, *others) if others else iterable_or_value
    try:
        return max((abs(x) for x in iterable_or_value), default=default, key=key)
    except TypeError:
        return abs(iterable_or_value)


del (Iterable, Any, Type, List, Tuple, Iterator, Protocol, TypeVar, TYPE_CHECKING, Sequence,
     Callable, ParamSpec, SupportsAbs)
del warnings, annotations



"""This module contains utilities pertaining to quantum information"""
from typing import Callable

import numpy as np
from scipy.spatial import transform as st

from . import linalg
from .linalg import paulis

__all__ = ['paulis', 'hadamard', 'cnot', 'cphase', 'crot', 'swap']


def crot(phi: float) -> np.ndarray:
    """conditional rotation by phi"""
    gate = np.array(
        [[1, 0, 0, 0],
         [0, 1, 0, 0],
         [0, 0, 1, 0],
         [0, 0, 0, np.exp(1j*phi)]]
    )
    return gate


hadamard = np.array(
    [[1, 1],
     [1, -1]]
)/np.sqrt(2)

cnot = np.array(
    [[1, 0, 0, 0],
     [0, 1, 0, 0],
     [0, 0, 0, 1],
     [0, 0, 1, 0]]
)

cphase = np.array(
    [[1, 0, 0, 0],
     [0, 1, 0, 0],
     [0, 0, 1, 0],
     [0, 0, 0, -1]]
)

swap = np.array(
    [[1, 0, 0, 0],
     [0, 0, 1, 0],
     [0, 1, 0, 0],
     [0, 0, 0, 1]]
)


def pauli_basis(n: int):
    """n-qubit normalized Pauli basis.

    Parameters
    ----------
    n: int
        Number of qubits.

    Returns
    -------
    sigma: ndarray, shape (d**2, d, d)
        n-qubit Pauli basis.

    """
    normalization = np.sqrt(2**n)
    combinations = np.indices((4,)*n).reshape(n, 4**n)
    sigma = linalg.tensor(*np.array(paulis)[combinations], rank=2)
    sigma /= normalization
    return sigma


def ggm_basis(d: int):
    """Hermitian orthonormal basis for d-dimensional Hilbert space.

    Parameters
    ----------
    d: int
        Dimension of Hilbert space.

    Returns
    -------
    Lambda: ndarray, shape (d**2, d, d)
        The basis (see [Bert08]_).

    References
    ----------
    .. [Bert08]
        Bertlmann, R. A., & Krammer, P. (2008). Bloch vectors for
        qudits. Journal of Physics A: Mathematical and Theoretical,
        41(23). https://doi.org/10.1088/1751-8113/41/23/235303
    """
    n_sym = int(d*(d - 1)/2)
    sym_rng = np.arange(1, n_sym + 1)
    diag_rng = np.arange(1, d)

    # Indices for offdiagonal elements
    j = np.repeat(np.arange(d - 1), np.arange(d - 1, 0, -1))
    k = np.arange(1, n_sym+1) - (j*(2*d - j - 3)/2).astype(int)
    j_offdiag = tuple(j)
    k_offdiag = tuple(k)
    # Indices for diagonal elements
    j_diag = tuple(i for m in range(d) for i in range(m))
    l_diag = tuple(i for i in range(1, d))

    inv_sqrt2 = 1/np.sqrt(2)
    Lambda = np.zeros((d**2, d, d), dtype=complex)
    Lambda[0] = np.eye(d)/np.sqrt(d)
    # First n matrices are symmetric
    Lambda[sym_rng, j_offdiag, k_offdiag] = inv_sqrt2
    Lambda[sym_rng, k_offdiag, j_offdiag] = inv_sqrt2
    # Second n matrices are antisymmetric
    Lambda[sym_rng+n_sym, j_offdiag, k_offdiag] = -1j*inv_sqrt2
    Lambda[sym_rng+n_sym, k_offdiag, j_offdiag] = 1j*inv_sqrt2
    # Remaining matrices have entries on the diagonal only
    Lambda[np.repeat(diag_rng, diag_rng)+2*n_sym, j_diag, j_diag] = 1
    Lambda[diag_rng + 2*n_sym, l_diag, l_diag] = -diag_rng
    # Normalize
    Lambda[2*n_sym + 1:, range(d), range(d)] /= np.tile(
        np.sqrt(diag_rng*(diag_rng + 1))[:, None], (1, d)
    )

    return Lambda


def unitary_to_rotation(U: np.ndarray) -> st.Rotation:
    """Convert a unitary operator to a scipy Rotation instance."""
    PTM = channel_to_liouville(apply_channel_unitary, pauli_basis(1), U)
    return st.Rotation.from_matrix(PTM[1:, 1:])


def apply_channel_unitary(rho, U):
    """Apply a unitary quantum channel to an input state rho."""
    return U @ rho @ U.conj().swapaxes(-2, -1)


def apply_channel_kraus(rho, K):
    """Apply the Kraus representation of a channel to the operator rho.

    Parameters
    ----------
    rho: ndarray, shape (d, d)
        The density operator going through the channel.
    K: ndarray, shape (..., n, d, d)
        The n Kraus operators of the channel.

    Returns
    -------
    rhop: ndarray, shape (d, d)
        The output density operator of the channel.

    """
    return (K @ rho @ K.conj().swapaxes(-2, -1)).sum(axis=-3)


def apply_channel_chi(rho, chi):
    """Apply the Chi matrix representation of a channel to the operator rho.

    Parameters
    ----------
    rho: ndarray, shape (d, d)
        The density operator going through the channel.
    chi: ndarray, shape (d**2, d**2)
        The chi matrix of the channel.

    Returns
    -------
    rhop: ndarray, shape (d, d)
        The output density operator of the channel.

    """
    n = int(np.log2(rho.shape[0]))
    sigma = pauli_basis(n)
    rhop = np.einsum('jk,jab,bc,kcd', chi, sigma, rho, sigma)
    return rhop


def apply_channel_liouville(rho, U, basis):
    """Apply the Liouville representation of a channel U to the operator rho.

    Parameters
    ----------
    rho: ndarray, shape (d, d)
        The density operator going through the channel.
    U: ndarray, shape (d**2, d**2)
        The Liouville representation of the channel.
    basis: ndarray, shape (d**2, d, d)
        The operator basis defining the Liouville representation.

    Returns
    -------
    rhop: ndarray, shape (d, d)
        The output density operator of the channel.

    """
    rhoket = np.tensordot(rho, basis, axes=[(-2, -1), (-1, -2)])
    rhoketp = U @ rhoket
    rhop = np.tensordot(rhoketp, basis, axes=[-1, 0])
    return rhop


def kraus_to_chi(K):
    """Convert from Kraus to chi matrix representation.

    Parameters
    ----------
    K: ndarray, shape (..., d, d)
        The Kraus operators.

    Returns
    -------
    chi: ndarray, shape (d**2, d**2)
        The chi (process) matrix.

    """
    n = int(np.log2(K.shape[-1]))
    sigma = pauli_basis(n)
    a_ij = np.einsum('iab,jba', K, sigma)
    chi = (a_ij.conj().T @ a_ij).T
    return chi


def convert_chi_liouville(chi_or_liouville, basis):
    """Convert from chi to Liouville representation and back.

    Parameters
    ----------
    chi_or_liouville: ndarray, shape (d**2, d**2)
        The chi (process) matrix.
    basis: ndarray, shape (d**2, d, d)
        The operator basis defining the Liouville / chi representations.

    Returns
    -------
    U: ndarray, shape (d**2, d**2)
        The converted channel.

    """
    traces = np.einsum('iab,jbc,kcd,lda', *[basis]*4)
    U = np.einsum('kl,ikjl', chi_or_liouville, traces)
    return U


def kraus_to_liouville(K, basis):
    """Convert from Kraus to Liouville representation.

    Parameters
    ----------
    K: ndarray, shape (..., d, d)
        The Kraus operators.
    basis: ndarray, shape (d**2, d, d)
        The operator basis defining the Liouville representation.

    Returns
    -------
    U: ndarray, shape (d**2, d**2)
        The Liouville representation.

    """
    U = np.einsum('iab,kbc,jcd,kad', basis, K, basis, K.conj())
    return U


def chi_to_choi(superoperator: np.ndarray, basis: np.ndarray):
    r"""Transform from Chi matrix to Choi matrix representation.

    Parameters
    ----------
    superoperator: array_like, shape (..., d**2, d**2)
        The superoperator to be transformed.
    basis: ndarray, shape (d**2, d, d)
        The basis defining the representation.

    Returns
    -------
    transformed: ndarray, shape (..., d**2, d**2)
        The transformed superoperator.

    Notes
    -----
    The transformation is given by

    .. math::

        \mathrm{choi}(\chi) = \mathcal{C}^T\chi\mathcal{C}

    with :math:`\mathcal{C}` the basis reshaped as a :math:`d^2\times
    d^2` matrix.
    """
    C = basis.reshape(superoperator.shape)
    choi = C.T @ superoperator @ C
    return choi


def choi_to_kraus(choi: np.ndarray):
    r"""Compute Kraus operators for a given Choi matrix.

    Parameters
    ----------
    choi: ndarray, shape (..., d**2, d**2)
        The choi matrix representation of a completely positive quantum
        map.

    Returns
    -------
    kraus: ndarray, shape (..., d**2, d, d)
        The Kraus operators of the quantum channel (completely positive
        map) represented by choi.

    Notes
    -----
    The Kraus (operator-sum) representation of a quantum channel
    :math:`\mathcal{S}` is given by

    .. math::

        \mathcal{S}(\bullet) = \sum_i A_i\bullet A_i^\dagger.

    The Kraus operators are obtained from the eigendecomposition of the
    Choi operator through the isomorphism

    .. math::

        A_i\simeq\sqrt{\lambda_i} v_i

    where :math:`\lambda_i, v_i` are the eigenvalues and -vectors of
    the Choi operator, respectively.
    """
    d2 = choi.shape[-1]
    d = int(np.sqrt(d2))
    outshape = choi.shape[:-1] + (d, d)

    eigvals, eigvecs = np.linalg.eigh(choi)
    kraus = np.sqrt(eigvals)[..., None, None] * eigvecs.T.reshape(outshape, order='F')
    return kraus


def choi_to_generalized_kraus(choi: np.ndarray):
    r"""Compute generalized Kraus operators for a given Choi matrix.

    Parameters
    ----------
    choi: ndarray, shape (..., d**2, d**2)
        The choi matrix representation of a (not necessarily completely
        positive) quantum map.

    Returns
    -------
    kraus_left, kraus_right: ndarray, shapes (d**2, d, d)
        The left and right generalized Kraus operators.

    Notes
    -----
    The map represented by the choi matrix
    :math:`\mathrm{choi}(\mathcal{S})` is decomposed into left and right
    generalized Kraus operators as

    .. math::

        \mathrm{choi}(\mathcal{S}) = \sum_i A_i\bullet B_i^\dagger,

    by decomposing :math:`\mathrm{choi}(\mathcal{S}) = U\Sigma
    V^\dagger` using singular value decomposition. Then there exists an
    ismorophism between left and right generalized Kraus operators and
    the SVD:

    .. math::

        A_i &\simeq \sqrt{\Sigma_i} U_i \\
        B_i &\simeq \sqrt{\Sigma_i} V_i.
    """
    d2 = choi.shape[-1]
    d = int(np.sqrt(d2))
    outshape = choi.shape[:-1] + (d, d)

    U, S, Vh = np.linalg.svd(choi, hermitian=True)

    sqrt_S = np.sqrt(S)
    kraus_left = sqrt_S[..., None, None] * U.T.reshape(outshape, order='F')
    kraus_right = sqrt_S[..., None, None] * Vh.conj().reshape(outshape, order='F')

    return kraus_left, kraus_right


def choi_to_liouville(superoperator: np.ndarray, basis: np.ndarray):
    r"""Transform from Choi matrix to Liouville representation.

    Parameters
    ----------
    superoperator: array_like, shape (..., d**2, d**2)
        The superoperator to be transformed.
    basis: Basis, shape (d**2, d, d)
        The basis defining the representation.

    Returns
    -------
    transformed: ndarray, shape (..., d**2, d**2)
        The transformed superoperator.

    Notes
    -----
    The transformation is given by

    .. math::

        \mathcal{S}_{ij} = \mathrm{tr}\left[\mathrm{choi}(\mathcal{S})
                                            (C_j^T\otimes C_i)\right]

    with :math:`C_i` a basis element.
    """
    liouville = np.empty(superoperator.shape, superoperator.dtype)
    for j in range(liouville.shape[-1]):
        liouville[..., j] = np.einsum('...ab,iba', superoperator, linalg.tensor(basis[j].T, basis))

    return liouville


def liouville_to_choi(U, basis):
    """Convert from Liouville to Choi matrix representation.

    Parameters
    ----------
    U: ndarray, shape (d**2, d**2)
        The Liouville representation.
    basis: ndarray, shape (d**2, d, d)
        The operator basis defining the Liouville representation.

    Returns
    -------
    choi: ndarray, shape (d**2, d**2)
        The Choi matrix of the channel U.

    """
    d2 = U.shape[0]

    try:
        basis_kroned = np.einsum(
            'iab,jcd->ijacbd', basis.transpose(0, 2, 1), basis
        ).reshape(d2, d2, d2, d2)
        choi = np.tensordot(U, basis_kroned, axes=[(0, 1), (1, 0)])
    except MemoryError:
        basis_kroned = np.empty((d2, d2, d2), dtype=complex)
        choi = np.empty((d2, d2), dtype=complex)
        for i in range(d2):
            basis_kroned = np.einsum(
                'ab,jcd->ijacbd', basis[i].transpose(0, 2, 1), basis,
                out=basis_kroned
            ).reshape(d2, d2, d2)
            choi[i] = np.tensordot(U, basis_kroned, axes=[(0, 1), (1, 0)])

    choi /= np.sqrt(d2)
    return choi


def channel_to_choi(U: Callable, d: int, *args, **kwargs):
    """Convert a channel U to the Choi representation.

    Parameters
    ----------
    U: Callable
        Callable that takes an operator as input and returns another operator.
    d: int
        Dimension of the Hilbert space.
    *args, **kwargs: Additional arguments passed through to U.

    Returns
    -------
    choi: ndarray, shape (d**2, d**2)
        The Choi matrix of the channel U.

    """
    eye = np.eye(d)
    choi = np.zeros((d**2, d**2), dtype=complex)
    for i in range(d):
        for j in range(d):
            ij = linalg.tensor(eye[:, i:i+1], eye[j:j+1])
            choi += linalg.tensor(ij, U(ij, *args, **kwargs))

    choi /= d
    return choi


def channel_to_liouville(U: Callable, basis: np.ndarray, *args, **kwargs):
    """Convert a channel U to the Liouville representation.

    Parameters
    ----------
    U: Callable
        Callable that takes an operator as input and returns another operator.
    basis: ndarray, shape (d**2, d, d)
        Orthonormal basis for the Hilbert space.
    *args, **kwargs: Additional arguments passed through to U.

    Returns
    -------
    choi: ndarray, shape (d**2, d**2)
        The Liouville representation matrix of the channel U.

    """
    try:
        basis_through_channel = U(basis, *args, **kwargs)
    except Exception:
        # Maybe broadcasting doesn't work with U
        try:
            first = U(basis[0], *args, **kwargs)
            basis_through_channel = np.zeros(first.shape[:-2] + basis.shape, basis.dtype)
            basis_through_channel[..., 0, :, :] = first
            for i, C in enumerate(basis[1:], start=1):
                basis_through_channel[..., i, :, :] = U(C, *args, **kwargs)
        except Exception as error:
            raise ValueError("Couldn't apply channel to basis") from error

    return np.einsum('aij,...bji->...ab', basis, basis_through_channel).real



import lazy_loader

__version__ = '2025.2.1'

"""Lazy imports as described in https://scientific-python.org/specs/spec-0001"""
__getattr__, __dir__, __all__ = lazy_loader.attach_stub(__name__, __file__)

del lazy_loader



from __future__ import annotations

import os
import pathlib
from typing import Optional, Union, Literal

import numpy as np
import tifffile
from moviepy.video.io.ImageSequenceClip import ImageSequenceClip

from qutil import io

try:
    from numpy.typing import NDArray
except ImportError:
    NDArray = np.ndarray

_pathT = Union[str, os.PathLike]


def convert_tiff(file: _pathT, fps: float, format: str = '.mp4', threads: Optional[int] = None,
                 out: Optional[_pathT] = None, logger: Optional[Literal['bar']] = None):
    """Converts a multipage .tif file to `format` and writes it to disk.

    Parameters
    ----------
    file : str | os.PathLike
        The .tif file to be converted.
    fps : float
        The framerate of the video.
    format : str
        The output format.
    threads : int, optional
        The number of CPU threads to use for the conversion (see
        :meth:`ImageSequenceClip.write_videofile`) and (possible)
        decompression (see :meth:`tifffile.TiffFile.asarray`)
    out : str | os.PathLike, optional
        A custom filename to write to. Otherwise, the original filename
        with a different suffix is used.
    logger : 'bar' | Proglog logger, optional
        See :meth:`ImageSequenceClip.write_videofile`.

    Examples
    --------
    `fps.txt` holds filenames and corresponding fps values::

        import pandas as pd
        import tempfile
        path = pathlib.Path('//janeway/User AG Bluhm/Common/GaAs',
                            'Hangleiter/characterization/vibrations/videos')
        out = pathlib.Path(tempfile.tempdir, 'foobar')
        tbl = pd.read_table(path / 'fps.txt',
                            dtype={'file': str, 'fps': float},
                            delim_whitespace=True)
        for _, series in tbl.iterrows():
            convert_tiff(path / series.file, series.fps, out=out, logger='bar',
                         threads=os.cpu_count())

    """
    in_name = pathlib.Path(file).with_suffix('.tif')
    out_name = io.query_overwrite((out or in_name).with_suffix(format))

    data = tifffile.imread(str(in_name), maxworkers=threads)
    # Grayscale tiffs returned without color channel in last axis,
    # but ImageSequenceClip implicitly expects it, so expand to 4
    # dims and broadcast the last axis to 3 (rgb).
    data = data.reshape(data.shape + (1,)*(4 - data.ndim))
    data = np.broadcast_to(data, data.shape[:3] + (3,))
    clip = ImageSequenceClip(list(data), fps=fps)

    clip.write_videofile(str(out_name), fps, audio=False, threads=threads,
                         logger=logger)



"""Programming tools"""
import pickle
import shelve
from typing import Callable, Mapping, Optional, Any
import inspect
import tempfile
import os.path
import functools
import pathlib
import dbm

import numpy as np

__all__ = ["file_cache", "lru_cache"]


def to_str_key(obj: Any) -> str:
    """Convert to a string representation that is unique except for
    - lists and ndarrays are treated the same

    :param obj:
    :return:
    """
    if isinstance(obj, tuple):
        return f'({",".join(map(to_str_key, obj))})'
    if isinstance(obj, np.ndarray):
        return to_str_key(obj.tolist())
    elif isinstance(obj, list):
        return f'[{",".join(map(to_str_key, obj))}]'
    elif isinstance(obj, Mapping):
        return '{%s}' % ",".join(f"{to_str_key(key)}: {to_str_key(value)}" for key, value in obj.items())
    elif isinstance(obj, (set, frozenset)):
        return f'{{{",".join(sorted(map(to_str_key, obj)))}}}'
    elif isinstance(obj, (int, float, complex, str, bytes, pathlib.Path)) or obj is None:
        return repr(obj)
    else:
        try:
            if eval(repr(obj)) == obj:
                return repr(obj)
        except RuntimeError:
            pass
        raise TypeError('not handled: ', type(obj))


class _AlwaysOpenCacheWrapper:
    def __init__(self, func: callable, db: shelve.Shelf):
        self.func = func
        self.db = db

    def __call__(self, *args, **kwargs):
        key = to_str_key((args, kwargs))
        if key in self.db:
            result = self.db[key]
        else:
            self.db[key] = result = self.func(*args, **kwargs)
        return result


class CachingWrapper:
    """This object wraps a callable and caches the results in a dbm database on the file system (by default in the temp
    folder). The key is generated via to_key which means that large arguments need a large time to process.

    >>> @file_cache
    ... def my_expensive_fn(a, b):
    ...     # do expensive stuff here
    ...     for x in range(a + b):
    ...         for y in range(x):
    ...             if x * y == a + b:
    ...                return x, y

    >>> my_expensive_fn(10, 5) # does calculation and saves result into default temporary folder
    (5, 3)

    >>> my_expensive_fn(10, 5) # loads result from disk
    (5, 3)

    The returned function can be used as a context manager to avoid reopening and closing the storage backend between
    calls.

    >>> large_list = [(1, 2), (3, 4)]
    >>> with my_expensive_fn as fn:
    ...     results = [fn(x, y) for x, y in large_list]

    This also works if the call is nested

    >>> def my_other_function(c):
    ...     return my_expensive_fn(c, c)

    >>> with my_expensive_fn:
    ...     results = [my_other_function(z) for z in range(1000)]

    The context manager is reentrant.
    """

    DEFAULT_ROOT = os.path.join(tempfile.gettempdir(), 'qutil_cache')
    PICKLE_PROTOCOL = pickle.HIGHEST_PROTOCOL

    def __init__(self, func, storage_path=None):
        self._func = func

        if storage_path is None:
            storage_path = os.path.join(self.DEFAULT_ROOT, self.get_full_function_name(func))

        self.storage_path = storage_path
        self._db: Optional[shelve.Shelf] = None
        self._wrapper = None
        self._entrance_counter = 0

    def __call__(self, *args, **kwargs):
        with self as wrapper:
            return wrapper(*args, **kwargs)

    def clear(self):
        """Clears cache of this function."""
        with self as wrapper:
            wrapper.db.clear()
    
    def __enter__(self) -> callable:
        if self._entrance_counter:
            assert self._db is not None
            assert self._wrapper is not None
        else:
            assert self._db is None
            assert self._wrapper is None
            folder = pathlib.Path(self.storage_path).parent
            if not folder.exists():
                folder.mkdir()
            self._db = shelve.open(self.storage_path, protocol=self.PICKLE_PROTOCOL)
            self._wrapper = _AlwaysOpenCacheWrapper(db=self._db, func=self._func)
        self._entrance_counter += 1
        return self._wrapper

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert self._entrance_counter > 0
        self._entrance_counter -= 1
        if self._entrance_counter == 0:
            self._db.close()
            self._db = None
            self._wrapper = None

    @classmethod
    def get_full_function_name(cls, func) -> str:
        return f"{inspect.getmodule(func).__name__}.{func.__name__}"

    @classmethod
    def clear_all_default_caches(cls):
        """Clear all caches that are in the default cache directory"""
        root = pathlib.Path(cls.DEFAULT_ROOT)
        if root.exists():
            for bak in root.glob('*.bak'):
                base_name = bak.with_suffix('')

                try:
                    shelve.open(str(base_name), 'r')
                except dbm.error:
                    print('Cannot open', bak.stem, 'and will not delete.')
                else:
                    bak.unlink()
                    base_name.with_suffix('.dir').unlink(missing_ok=True)
                    base_name.with_suffix('.dat').unlink(missing_ok=True)


def file_cache(func: Callable) -> Callable:
    """See doc of CachingWrapper."""
    return CachingWrapper(func)


lru_cache = functools.lru_cache



"""Exposes fastz under a nicer name."""
import fastz
import numpy as np
import importlib

_REEXPORTED_CLASSES = [
    'R',
    'C',
    'L',
    'Z',
    'SeriesZ',
    'ParallelZ',
    'LumpedElement',
]

# we do this instead of 'from tikz import ...' to
#  1. avoid duplication with __all__
#  2. make this code work with a mocked fastz
for cls in _REEXPORTED_CLASSES:
    locals()[cls] = getattr(fastz, cls)


__all__ = [
    'fastz',
    'Rv',
] + _REEXPORTED_CLASSES


class Rv(R):
    """Virtual AC input resistance of a TIA."""

    @property
    def prefix(self):
        return 'Rv'

    def __call__(self, ff, **lumpedparam):
        return np.asarray(ff) * self._lookup_value(**lumpedparam)


del cls
del importlib



"""This is the new name of the qutil module which is just an alias to qutil for backwards compatibility reasons."""
import sys
import lazy_loader

import qutil

__version__ = qutil.__version__
__all__ = qutil.__all__

for module_name in qutil.__all__:
    globals()[module_name] = lazy_loader.load(f'qutil.{module_name}')

sys.modules[__name__] = qutil

del qutil

del lazy_loader
del sys



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 25 11:14:02 2022

@author: Hangleiter
"""
import pathlib
import warnings
from itertools import compress
from typing import Dict, Union

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

try:
    import ezdxf
    from ezdxf import recover
    from ezdxf.addons.drawing import RenderContext, Frontend
    from ezdxf.addons.drawing.properties import LayoutProperties
    from ezdxf.addons.drawing.matplotlib import MatplotlibBackend
except ImportError:
    ezdxf = None


DEFAULT_TEXT_KWARGS = dict(backgroundcolor='black',
                           horizontalalignment='center',
                           verticalalignment='center',
                           color='white', fontsize=8)



class GateLayout:
    def __init__(self, layout_file=None, gate_names=None, gate_mask=None,
                 background_color='#ffffff', foreground_color='tab:gray', cmap='hot', v_min=-2,
                 v_max=0, explode_factor=0.1, fignum=998, text_kwargs=None, offset=(0., 0.)):

        if ezdxf is None:
            raise RuntimeError('Could not import ezdxf which is required to read DXF files.')

        if layout_file is None:
            layout_file = (pathlib.Path(r'\\janeway\User AG Bluhm\Hangleiter\Devices\gl_005d')
                           / 'gl_005d.dxf')

        if gate_names is None:
            gate_names = ['LT', 'LP', 'LB', 'PA', 'TAB', 'PB', 'TBC', 'PC', 'TCD', 'PD', 'RB',
                          'RP', 'RT', 'SD', 'RFD', 'NCD', 'RFC', 'NBC', 'RFB', 'NAB', 'RFA', 'SA']

        if gate_mask is None:
            gate_mask = [True]*14 + [False, True, False, True, False, True, False, True]

        if len(gate_names) != len(gate_mask):
            warnings.warn(f"Gate name number is different from gate mask number: {len(gate_names)} != {len(gate_mask)}",
                          stacklevel=2)

        self.fig = plt.figure(fignum)
        self.ax = self.fig.add_axes([0, 0, 1, 1])
        self.layout_file = layout_file
        self.gate_names = gate_names
        self.gate_mask = gate_mask
        self.explode_factor = explode_factor

        self.background_color = matplotlib.colors.to_hex(background_color)
        self.foreground_color = matplotlib.colors.to_hex(foreground_color)
        self.cmap = matplotlib.cm.get_cmap(cmap)
        self.norm = matplotlib.colors.Normalize(v_min, v_max)
        self.v_min = self.norm.vmin
        self.v_max = self.norm.vmax
        self.text_kwargs = {**DEFAULT_TEXT_KWARGS, **(text_kwargs or {})}

        self._latest_voltages = np.zeros(len(self.gate_names))
        self._offset = np.array(offset)

        self.patch_collection, self.patches, self.texts = self._setup_figure()

    def __call__(self):
        self.update()

    def __repr__(self):
        self.__call__()
        return super().__repr__()

    def update(self, voltages: Union[np.ndarray, Dict[str, float]] = None, force: bool = False):
        """Update the displayed voltages.

        Parameters
        ----------
        voltages: Union[np.ndarray, Mapping[str, float]], optional
            Either an array_like with all voltages that are displayed,
            or a dict with structure {gate_name: voltage}. If None,
            voltages are obtained from :meth:`get_voltages` that can be
            overridden by subclasses.
        force: bool, optional
            Force an update by querying the instruments. Otherwise
            values are obtained from cache. Passed along to
            :meth:`get_voltages`.

        """
        # Masked gates are displayed as 0.
        if voltages is None:
            self.get_voltages(force)
        elif isinstance(voltages, dict):
            for gate, voltage in voltages.items():
                self._latest_voltages[self.gate_names.index(gate)] = voltage
        else:
            self._latest_voltages[:] = np.broadcast_to(voltages, self._latest_voltages.shape)

        # Here we update the voltages
        self.patch_collection.set_array(self._latest_voltages)

        # Here we update the texts
        for i, (text, gate, volt) in enumerate(
                compress(zip(self.texts, self.gate_names, self._latest_voltages), self.gate_mask)
        ):
            text.set_text(f'{i}: {gate}\n{volt:.3f}')

    def get_voltages(self, force: bool = False):
        return self._latest_voltages

    def _setup_figure(self):
        vertices = get_vertices_from_dfx(self.layout_file, self.background_color, self.foreground_color)
        if len(self.gate_names) != len(vertices):
            warnings.warn(f"Found {len(vertices)} patches in file but got {len(self.gate_names)} gate names", stacklevel=2)

        patches = []
        texts = []
        for idx, verts in enumerate(vertices):
            verts += self._offset[None, :]
            patches.append(matplotlib.patches.Polygon(verts, True, zorder=0))
            texts.append(matplotlib.text.Text(*verts.mean(axis=0), f'INIT', **self.text_kwargs))

        all_coords = np.array([(patch.get_xy().mean(axis=0)) for patch in patches])

        # Order everything by angle around the center of mass
        angles = np.angle((all_coords - all_coords.mean(axis=0)).view(complex))
        texts = [text for _, text in sorted(zip(angles.flat, texts))]
        patches = [patch for _, patch in sorted(zip(angles.flat, patches))]

        # redraw everything
        self.ax.cla()

        pc = matplotlib.collections.PatchCollection(patches, norm=self.norm, cmap=self.cmap)
        pc.set_edgecolors('tab:gray')
        self.ax.add_collection(pc)

        for i, txt in enumerate(texts):
            # shift every other text by given percentage up or down to decrowd
            xpos, ypos = txt.get_position()
            txt.set_y(ypos*(1 + self.explode_factor*(-1)**(i % 2)))

            try:
                name = self.gate_names[i]
            except IndexError:
                name = f'Unknown {i}'

            txt.set_text(f'{name}: NaN')
            self.ax.add_artist(txt)
        self.ax.autoscale_view(True)
        return pc, patches, texts


def get_vertices_from_dfx(file_name, background_color, foreground_color, vert_drop=0.01):
    # Safe loading procedure (requires ezdxf v0.14):
    try:
        doc, auditor = recover.readfile(file_name)
    except IOError as ioe:
        raise IOError('Not a DXF file or a generic I/O error.') from ioe
    except ezdxf.DXFStructureError as e:
        raise RuntimeError('Invalid or corrupted DXF file.') from e

    tmp_fig = plt.figure(visible=False)
    tmp_ax = tmp_fig.add_axes([0, 0, 1, 1])
    tmp_fig.set_visible(False)

    vertices = []

    # ezdxf is very object-bloated. No clue how to get to coordinates.
    # Therefore we extract vertices from drawn artists
    try:
        ctx = RenderContext(doc)
        out = MatplotlibBackend(tmp_ax)
        Frontend(ctx, out).draw_layout(
            doc.modelspace(),
            finalize=True,
            layout_properties=LayoutProperties('Ebeam',
                                               background_color,
                                               foreground_color))

        # apply trafo if necessary
        transform = tmp_ax.transData.inverted().transform
        for child in tmp_ax.get_children():
            if isinstance(child, matplotlib.patches.PathPatch):
                verts = transform(child.get_verts())
                # Drop vertices that are less than 1% apart
                # Last vertex is the first
                verts = np.vstack(
                    [verts[1:][np.linalg.norm(np.diff(verts, axis=0), axis=1) >= vert_drop],
                     verts[-1]]
                )
                vertices.append(verts)
    finally:
        plt.close(tmp_fig)
    return vertices



import operator
import string
from functools import reduce
from itertools import zip_longest
from typing import List, NamedTuple, Sequence, Tuple, Union

import numpy as np
from numpy import linalg, ndarray
from scipy.linalg import polar

try:
    import numba as nb

    numba = True
except ImportError:
    numba = False


def abs2(x, /, out=None, *, where=True, casting='same_kind', order='C', dtype=None) -> ndarray:
    r"""
    Fast absolute value squared,

    .. math::

        |\cdot|^2 = \Re(\cdot)^2 + \Im(\cdot)^2.

    Equivalent to::

        abs(x)**2.

    See :class:`numpy:numpy.ufunc` and the `NumPy reference`_ for
    documentation of the arguments.

    .. _NumPy reference: https://numpy.org/doc/stable/reference/ufuncs.html

    Examples
    --------
    >>> abs2([(1 + 1j) / np.sqrt(2)])
    array([1.])
    >>> np.abs([(1 + 1j) / np.sqrt(2)])**2
    array([1.])

    """
    x = np.asanyarray(x)
    out = np.empty(x.shape, order=order, dtype=dtype or np.float64) if out is None else out
    out = np.square(x.real, out=out, dtype=dtype, where=where, casting=casting)
    if np.iscomplexobj(x):
        tmp = np.square(x.imag, order=order, dtype=dtype, where=where, casting=casting)
        out = np.add(out, tmp, out=out, dtype=dtype, where=where, casting=casting)
    return out


def cexp(x, /, out=None, *, where=True, casting='same_kind', order='C', dtype=None) -> ndarray:
    r"""
    Fast complex exponential,

    .. math::

        \exp(i x).

    See :class:`numpy:numpy.ufunc` and the `NumPy reference`_ for
    documentation of the arguments.

    .. _NumPy reference: https://numpy.org/doc/stable/reference/ufuncs.html

    Examples
    --------
    >>> cexp([np.pi])
    array([-1.+1.2246468e-16j])
    >>> np.exp([1j * np.pi])
    array([-1.+1.2246468e-16j])

    """
    x = np.asanyarray(x)
    out = np.empty(x.shape, order=order, dtype=dtype or np.complex128) if out is None else out
    out.real = np.cos(x, out=out.real, where=where, casting=casting)
    out.imag = np.sin(x, out=out.imag, where=where, casting=casting)
    return out


if numba:
    # nb.vectorize generates ufuncs with all the kwargs, so only one argument required.
    def _abs2(x):
        result = x.real * x.real
        if np.iscomplexobj(x):
            return result + x.imag * x.imag
        return result


    def _cexp(x):
        return np.cos(x) + 1j * np.sin(x)


    _abs2.__doc__ = abs2.__doc__
    _cexp.__doc__ = cexp.__doc__
    abs2 = nb.vectorize([nb.float32(nb.float32),
                         nb.float32(nb.complex64),
                         nb.float64(nb.float64),
                         nb.float64(nb.complex128)],
                        target='parallel',
                        cache=True)(_abs2)
    cexp = nb.vectorize([nb.complex64(nb.float32),
                         nb.complex128(nb.float64)],
                        target='parallel',
                        cache=True)(_cexp)


def remove_float_errors(arr: ndarray, eps_scale: float = None):
    """
    Clean up arr by removing floating point numbers smaller than the dtype's
    precision multiplied by eps_scale. Treats real and imaginary parts
    separately.
    """
    if eps_scale is None:
        atol = np.finfo(arr.dtype).eps * arr.shape[-1]
    else:
        atol = np.finfo(arr.dtype).eps * eps_scale

    # Hack around arr.imag sometimes not being writable
    if arr.dtype == complex:
        arr = arr.real + 1j * arr.imag
        arr.real[np.abs(arr.real) <= atol] = 0
        arr.imag[np.abs(arr.imag) <= atol] = 0
    else:
        arr = arr.real
        arr.real[np.abs(arr.real) <= atol] = 0

    return arr


def max_abs_diff(a, b):
    """Maximum absolute difference"""
    return np.max(np.abs(a - b))


def max_rel_diff(a, b):
    """Maximum relative difference"""
    return np.nanmax(np.abs((a - b) / np.abs(b)))



"""Functions and decorators to  help with runtime typechecking. Notably the `@check_literals` decorator to ensure that
arguments match an annotated literal. Wildcard imports the `typeguard` which provides the powerful `@typechecked`
decorator.
"""

import inspect
import typing
import functools

import typeguard
from typeguard import *


__all__ = [x for x in typeguard.__dict__ if not x.startswith('_')] + ["check_literals"]


def _get_annotated_literals(signature: inspect.Signature) -> typing.Mapping[str, tuple]:
    literals = {}
    for name, parameter in signature.parameters.items():
        annotation = parameter.annotation
        additional_values = ()

        # check for Optional[Literal[...]]
        if typing.get_origin(annotation) == typing.Union:
            union_types = typing.get_args(annotation)
            if len(union_types) == 2 and type(None) in union_types:
                annotation, = set(union_types) - {type(None)}
                additional_values = (None,)

        if typing.get_origin(annotation) == typing.Literal:
            literals[name] = typing.get_args(annotation) + additional_values
    return literals


def check_literals(function: callable) -> callable:
    """A decorator that checks if arguments are valid for all typing.Literal annotated function arguments.

    This decorator handles Literal[...] and Optional[Literal[...]].

    Examples
    --------
    >>> from typing import Literal, Sequence
    >>> @check_literals
    ... def my_function(a: Sequence[int], b: Literal['forward', 'backward']):
    ...     if b == 'backward':
    ...         return list(reversed(a))
    ...     else:
    ...         # b is guaranteed to be 'forward' here
    ...         return list(a)

    works

    >>> my_function([1, 2, 3], 'backward')
    [3, 2, 1]

    works because the first arguement is not checked at runtime

    >>> my_function({'no': 'sequence', 'yes': 'mapping'}, 'backward')
    ['yes', 'no']

    Here we get a runtime error because of typo in 'backward'

    >>> my_function('wrong', 'backwardd')
    Traceback (most recent call last):
    ...
    ValueError: The argument b of ...typecheck.my_function has to be in ('forward', 'backward') but was 'backwardd'.

    This decorator has an overhead per function call of several microseconds (~10µs) and roughly one order of
    magnitude less than ``typeguard.typechecked``.

    If you need runtime-typechecking for all arguments use the typeguard functions that are exposed in this module for
    your convenience.
    """

    signature = inspect.signature(function)
    parameters = signature.parameters
    literals = _get_annotated_literals(signature)
    var_positional_name = next((name
                                for name, param in parameters.items()
                                if param.kind == inspect.Parameter.VAR_POSITIONAL),
                               None)
    var_keyword_name = next((name
                             for name, param in parameters.items()
                             if param.kind == inspect.Parameter.VAR_KEYWORD),
                            None)

    function_name = f'{inspect.getmodule(function).__name__}.{function.__name__}'

    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        bound = signature.bind(*args, **kwargs)
        for arg_name, arg in bound.arguments.items():
            values = literals.get(arg_name, None)
            if values is not None:
                if arg_name == var_positional_name:
                    for idx, value in enumerate(arg):
                        if value not in values:
                            raise ValueError(f"The positional argument {arg_name}[{idx}] of {function_name} has "
                                             f"to be in {values!r} but was {value!r}.")
                elif arg_name == var_keyword_name:
                    for key, value in arg.items():
                        if value not in values:
                            raise ValueError(f"The keyword argument {arg_name}[{key!r}] of {function_name} has "
                                             f"to be in {values!r} but was {value!r}.")
                elif arg not in values:
                    raise ValueError(f"The argument {arg_name} of {function_name} has "
                                     f"to be in {values!r} but was {arg!r}.")

        return function(*args, **kwargs)
    return wrapper



"""This module contains UI utility functions."""
from __future__ import annotations

import logging
import os
import socketserver
import sys
import webbrowser
from http.server import SimpleHTTPRequestHandler
from threading import Thread
from typing import Iterable, Optional, TextIO

from ..functools import wraps

__all__ = ['progressbar', 'progressbar_range']


def _in_spyder_kernel():
    """Determine if we are currently running inside spyder on a best effort basis."""
    return 'SPY_PYTHONPATH' in os.environ


def _in_notebook_kernel():
    # https://github.com/jupyterlab/jupyterlab/issues/16282
    return 'JPY_SESSION_NAME' in os.environ and os.environ['JPY_SESSION_NAME'].endswith('.ipynb')


def _in_jupyter_kernel():
    # https://discourse.jupyter.org/t/how-to-know-from-python-script-if-we-are-in-jupyterlab/23993
    return 'JPY_PARENT_PID' in os.environ


if not _in_notebook_kernel():
    if _in_jupyter_kernel():
        # Autonotebook gets confused in jupyter consoles
        from tqdm.std import tqdm, trange
    else:
        from tqdm.autonotebook import tqdm, trange
else:
    from tqdm.notebook import tqdm, trange


class TCPServerThread(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


class ThreadedWebserver:
    """Serves a simple HTTP server in a background thread.

    If *qcodes_monitor_mode* is true, the port will be grabbed from
    :mod:`qcodes:qcodes.monitor.monitor`.

    Examples
    --------
    Run as server for the qcodes monitor:

    >>> from qcodes.monitor import Monitor
    >>> from qcodes.parameters import ManualParameter
    >>> server = ThreadedWebserver(qcodes_monitor_mode=True)
    >>> server.show()
    >>> monitor = Monitor(ManualParameter('foobar'))

    Join the threads again:

    >>> monitor.stop()
    >>> server.stop()

    """

    def __init__(self, qcodes_monitor_mode: bool = False, url: str = 'localhost',
                 port: int = 3000):
        if qcodes_monitor_mode:
            # Copy stuff from qcodes.monitor.monitor. Purely convenience
            from qcodes.monitor import monitor
            self.port = monitor.SERVER_PORT
            self.log = monitor.log
            self.static_dir = STATIC_DIR = os.path.join(os.path.dirname(monitor.__file__), "dist")

            class HTTPRequestHandler(SimpleHTTPRequestHandler):
                def __init__(self, *args, directory: str | None = ..., **kwargs):
                    super().__init__(*args, directory=STATIC_DIR, **kwargs)

        else:
            self.port = port
            self.log = logging.getLogger(__name__)

            class HTTPRequestHandler(SimpleHTTPRequestHandler):
                ...
        self.url = url

        self.log.info(f"Starting HTTP Server at http://{self.url}:{self.port}")
        self.server = TCPServerThread(("", self.port), HTTPRequestHandler)
        self.thread = self.server.server_thread = Thread(target=self.server.serve_forever)
        self.thread.start()

    def __del__(self):
        self.stop()

    def stop(self):
        self.log.info("Shutting Down HTTP Server")
        self.server.shutdown()
        self.server.server_close()
        self.thread.join()

    def show(self, new=0, autoraise=True):
        """Show the server in a browser.

        See :func:`webbrowser:webrowser.open` for parameters.
        """
        webbrowser.open(f"http://{self.url}:{self.port}", new=new, autoraise=autoraise)

    open = show


def _simple_progressbar(iterable: Iterable, desc: str = "Computing", disable: bool = False,
                        total: Optional[int] = None, size: int = 25, file: TextIO = sys.stdout,
                        *_, **__):
    """https://stackoverflow.com/a/34482761"""
    if disable:
        yield from iterable
        return

    if total is None:
        try:
            total = len(iterable)
        except TypeError:
            raise ValueError(f'{iterable} has no len, please supply the total argument.')

    def show(j):
        x = int(size*j/total)
        file.write("\r{}:\t[{}{}] {} %".format(desc, "#"*x, "."*(size - x),
                   int(100*j/total)))
        file.flush()

    show(0)
    for i, item in enumerate(iterable):
        yield item
        show(i + 1)

    file.write("\n")
    file.flush()


class ProgressbarLock:
    """A global lock for progressbars used as a decorator.

    The intended use case is to keep nested progressbars from interfering in
    non-interactive (notebook) mode.

    Examples
    --------
    >>> import time
    >>> for i in progressbar_range(10, desc='outer', file=sys.stdout):
    ...     for j in progressbar_range(10, desc='inner', file=sys.stdout, leave=False):
    ...         time.sleep(0.05) # doctest: +NORMALIZE_WHITESPACE
    outer: ...
    >>> for i in progressbar_range(10, desc='outer', file=sys.stdout, disable=True):
    ...     for j in progressbar_range(10, desc='inner', file=sys.stdout):
    ...         time.sleep(0.05)  # doctest: +NORMALIZE_WHITESPACE
    inner: ...

    Does nothing on a single, not nested bar:

    >>> for i in progressbar('asdf', desc='only', file=sys.stdout):
    ...     time.sleep(0.05)  # doctest: +NORMALIZE_WHITESPACE
    only: ...

    """
    _stack: list[bool] = []
    DISABLED: bool = False
    """Globally disable the locking mechanism."""

    def __call__(self, func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            # No need to lock if the bar is disabled dynamically.
            disable = kwargs.get('disable', False)

            if not self.DISABLED and (disable or not all(self._stack)):
                self.lock(True)
            else:
                self.lock(False)

            kwargs['disable'] = self.locked
            try:
                yield from func(*args, **kwargs)
            finally:
                self.release()

        return wrapper

    @property
    def locked(self) -> bool:
        try:
            return self._stack[-1]
        except IndexError:
            return False

    def lock(self, val: bool):
        """Lock the bar in the current loop if True."""
        self._stack.append(val)

    def release(self):
        self._stack.pop()


def auto_progress_bar_lock(pbar_function: callable) -> callable:
    """Enable progress bar lock for the given progress bar if a qtconsole kernel is detected."""
    # The second condition should also fire if in a spyder kernel, so technically doppelt-gemoppelt
    if _in_spyder_kernel() or (_in_jupyter_kernel() and not _in_notebook_kernel()):
        return ProgressbarLock()(pbar_function)
    else:
        return pbar_function


@auto_progress_bar_lock
def progressbar(iterable: Iterable, *args, **kwargs):
    """
    Progress bar for loops. Uses tqdm if available or a quick-and-dirty
    implementation from stackoverflow.

    Usage::

        for i in progressbar(range(10)):
            do_something()

    See :class:`~tqdm.tqdm` or :func:`_simple_progressbar` for possible
    args and kwargs.
    """
    if tqdm is not None:
        return tqdm(iterable, *args, **kwargs)
    else:
        return _simple_progressbar(iterable, *args, **kwargs)


@auto_progress_bar_lock
def progressbar_range(*args, **kwargs):
    if tqdm is not None:
        return trange(*args, **kwargs)
    else:
        return _simple_progressbar(range(*args), **kwargs)



import threading
import time


class ThreadedPeriodicCallback:
    """Periodically call the given function. Dont forget to call start. You can change the period while it runs.
    Be aware that your callback needs to be thread safe. This means you need to make sure that the state there is always
    consistent.

    Example:
        >>> pcb = ThreadedPeriodicCallback(1., lambda: print('my function'))  # doctest: +SKIP
        >>> pcb.start()  # doctest: +SKIP

        >>> pcb.stop()  # doctest: +SKIP
        Stop the parallel thread
    """
    def __init__(self, period: float, callback: callable):
        self._stop = None
        self.period = period
        self.callback = callback
        self.thread = None

    def _run(self):
        while not self._stop:
            self.callback()
            time.sleep(self.period)

    def stop(self):
        if self.thread is not None:
            self._stop = True
            print('Stop the parallel thread')
            self.thread.join()
            self.thread = None

    def start(self):
        if self.thread is not None:
            raise RuntimeError("Thread is still running")

        self._stop = False
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def __del__(self):
        self.stop()



__all__ = ['progressbar', 'progressbar_range', 'ProgressbarLock', 'GateLayout', 'QcodesGateLayout',
           'ThreadedWebserver']

from .core import progressbar, progressbar_range, ProgressbarLock, ThreadedWebserver
from .gate_layout import GateLayout, QcodesGateLayout



from __future__ import annotations

import abc
import numbers
from abc import ABC
from dataclasses import dataclass, field, InitVar
from math import ceil, floor, inf, nan
from typing import (Callable, Container, Generic, Iterable, Protocol, TypeVar, Union,
                    runtime_checkable, Set)

from numpy import iterable as is_iterable

from .functools import cached_property
from .itertools import next_closest, next_largest, next_smallest

_RealT = TypeVar("_RealT", bound=numbers.Real)
_IntegralT = TypeVar("_IntegralT", bound=numbers.Integral)
_BoundT = Union[Callable[[], _RealT], _RealT]
_BoundT = Union[Iterable[_BoundT], _BoundT]


class Bound(Generic[_RealT]):
    """A lazily evaluated bound."""

    def __init__(self, default: _RealT, integral: bool):
        self.default = default
        self.integral = integral

    def __set_name__(self, owner, name):
        self._name = f'_{name}'

    def __get__(self, instance, owner=None) -> _RealT:
        if instance is None:
            return self.default

        value = getattr(instance, self._name, {self.default})
        evaluated = set()
        for item in value:
            if callable(item):
                evaluated.add(item())
            else:
                evaluated.add(item)
        if self._name.startswith('_lower'):
            lower = max(evaluated)
            if self.integral:
                try:
                    return self.cast_lower(lower)
                except (OverflowError, ValueError):
                    # inf or nan
                    return lower
            return lower
        if self._name.startswith('_upper'):
            upper = min(evaluated)
            if self.integral:
                try:
                    return self.cast_upper(upper)
                except (OverflowError, ValueError):
                    # inf or nan
                    return upper
            return upper
        raise NameError('Bound descriptor instance name should start with lower or upper')

    def __set__(self, instance, value: _BoundT):
        if value is self:
            value = self.default
        if not is_iterable(value):
            value = {value}
        else:
            value = set(value)
        setattr(instance, self._name, value)

    def cast_lower(self, lower):
        return ceil(lower)

    def cast_upper(self, upper):
        return floor(upper)


class ReciprocalBound(Bound[_RealT]):

    def cast_lower(self, lower):
        # 1/floor is less inclusive than 1/ceil
        return floor(lower)

    def cast_upper(self, upper):
        # 1/ceil is less inclusive than 1/floor
        return ceil(upper)


class Bounded(Protocol[_RealT]):
    """An object with bounds."""
    lower: Bound[_RealT]
    upper: Bound[_RealT]


@runtime_checkable
class Reciprocal(Protocol[_RealT]):
    """A reciprocal quantity that's specified by a numerator."""
    numerator: _RealT

    def _to_denominator(self, it: Iterable[_RealT | Callable[[], _RealT]]) -> Set[_RealT]:
        """Converts rationals in it to denominators given numerator."""
        result = set()
        for item in it:
            if callable(item):
                result.add(lambda: self.numerator / item())
            else:
                result.add(self.numerator / item)
        return result


class Domain(Container[_RealT], Bounded[_RealT]):
    """A domain that constrains the allowed values a parameter can take."""
    lower: Bound[_RealT]
    upper: Bound[_RealT]
    precision: int

    def __lt__(self, value: _RealT) -> bool:
        return self.round(self.max() - value) < 0

    def __gt__(self, value: _RealT) -> bool:
        return self.round(self.min() - value) > 0

    @abc.abstractmethod
    def __and__(self, other: object) -> _DomainT:
        ...

    def __contains__(self, value: object) -> bool:
        return isinstance(value, numbers.Real) and not (value > self or value < self)

    @abc.abstractmethod
    def min(self) -> _RealT:
        ...

    @abc.abstractmethod
    def max(self) -> _RealT:
        ...

    @abc.abstractmethod
    def next_closest(self, value: _RealT) -> _RealT:
        ...

    @abc.abstractmethod
    def next_smallest(self, value: _RealT) -> _RealT:
        ...

    @abc.abstractmethod
    def next_largest(self, value: _RealT) -> _RealT:
        ...

    def round(self, value: _RealT):
        if self.precision is None:
            return value
        try:
            # NumPy object, cast to builtin type to avoid differences
            # in the implementation of __round__
            return round(value.item(), self.precision)
        except AttributeError:
            return round(value, self.precision)


_DomainT = TypeVar('_DomainT', bound=Domain)


@dataclass
class Interval(Domain[_RealT], Bounded[_RealT], ABC):
    """An interval with lower and upper bounds."""
    lower: _BoundT[_RealT] = Bound[_RealT](default=-inf, integral=False)
    upper: _BoundT[_RealT] = Bound[_RealT](default=+inf, integral=False)
    precision: int = field(default=16, repr=False)

    def __and__(self, other: object) -> _DomainT:
        if isinstance(other, Domain) and isinstance(other, Reciprocal):
            # let reciprocal class handle it
            return other & self

        if isinstance(other, Interval):
            cls = type(self)
            return cls(lower=self._lower | other._lower,
                       upper=self._upper | other._upper,
                       precision=min(self.precision, other.precision))

        if isinstance(other, Iterable):
            if not isinstance(other, BoundedSet):
                other = BoundedSet(other)
            return other & self

        return NotImplemented

    def min(self) -> _RealT:
        return self.lower

    def max(self) -> _RealT:
        return self.upper


@dataclass
class DiscreteInterval(Interval[_IntegralT], Bounded[_IntegralT]):
    """A discrete interval that only allows integral values.

    Examples
    --------
    >>> x = 7
    >>> domain = DiscreteInterval(lower=-4.3, upper={5, lambda: x})
    >>> domain
    DiscreteInterval(lower=-4, upper=5)
    >>> 4 in domain
    True
    >>> 4.0 in domain  # only ints
    False
    >>> domain.next_closest(1.7)
    2
    >>> domain.next_smallest(1.7)
    1

    Bounds can be dynamic and are lazily evaluated:

    >>> x = 2
    >>> 4 in domain
    False
    >>> domain
    DiscreteInterval(lower=-4, upper=2)
    """
    lower: _BoundT[_IntegralT] = Bound[_IntegralT](default=-inf, integral=True)
    upper: _BoundT[_IntegralT] = Bound[_IntegralT](default=+inf, integral=True)

    def __contains__(self, value: object) -> bool:
        return isinstance(value, numbers.Integral) and super().__contains__(value)

    def to_bounded_set(self) -> BoundedSet:
        return BoundedSet(range(self.min(), self.max() + 1),
                          lower=self.min(),
                          upper=self.max(),
                          precision=self.precision)

    def next_closest(self, value: _RealT) -> int:
        if value < self:
            return ceil(self.min())
        if value > self:
            return floor(self.max())
        return round(value)

    def next_smallest(self, value: _RealT) -> int:
        if value < self:
            return ceil(self.min())
        if value > self:
            return floor(self.max())
        return floor(value)

    def next_largest(self, value: _RealT) -> int:
        if value < self:
            return ceil(self.min())
        if value > self:
            return floor(self.max())
        return ceil(value)


class ContinuousInterval(Interval[_RealT]):
    """A continuous interval that allows any value within its bounds.

    Examples
    --------
    >>> domain = ContinuousInterval(-3.1, 1.7)
    >>> 0 in domain
    True
    >>> 1.7 in domain
    True
    >>> 10 in domain
    False
    >>> -10 < domain
    True
    >>> domain.next_smallest(0.91)
    0.91
    >>> domain.next_largest(6)
    1.7

    """

    def next_closest(self, value: _RealT) -> _RealT:
        if value < self:
            return self.min()
        if value > self:
            return self.max()
        return value

    def next_smallest(self, value: _RealT) -> _RealT:
        return self.next_closest(value)

    def next_largest(self, value: _RealT) -> _RealT:
        return self.next_closest(value)


@dataclass
class ReciprocalDiscreteInterval(DiscreteInterval[_IntegralT], Reciprocal[_RealT]):
    """A rational interval that only allows integral denominators.

    Examples
    --------
    The interval [2.3/10, 2.3/9, ..., 2.3/1]:
    >>> domain = ReciprocalDiscreteInterval(lower=1, upper=10, numerator=2.3)
    >>> 2.3/8 in domain
    True
    >>> 2.4/8 in domain
    False
    >>> 2.3/11 < domain
    True
    >>> domain.next_smallest(0.3)  # 2.3/8
    0.2875
    >>> domain.next_largest(0.3)  # 2.3/7
    0.32857142857142857
    >>> domain.next_closest(0.3)  # 2.3/8
    0.2875

    """
    lower: _BoundT[_IntegralT] = ReciprocalBound[_IntegralT](default=1, integral=True)
    upper: _BoundT[_IntegralT] = ReciprocalBound[_IntegralT](default=inf, integral=True)
    numerator: _RealT = 1.0

    def __post_init__(self):
        self._denominator_interval = DiscreteInterval(self.lower, self.upper, self.precision)

    def __contains__(self, value: object) -> bool:
        # Works only without floating point imprecision
        denominator = self.numerator / value
        return not self.round(denominator % 1) and int(denominator) in self._denominator_interval

    def __and__(self, other: object) -> _DomainT:
        cls = type(self)
        if isinstance(other, DiscreteInterval):
            if isinstance(other, cls) and other.numerator == self.numerator:
                return cls(numerator=self.numerator,
                           lower=self._lower | other._lower,
                           upper=self._upper | other._upper,
                           precision=min(self.precision, other.precision))
            try:
                return self.to_bounded_set() & other.to_bounded_set()
            except TypeError:
                raise NotImplementedError(f'Intersection not implemented for {self} & {other}')
        if isinstance(other, Interval):
            return cls(numerator=self.numerator,
                       lower=self._lower | self._to_denominator(other._upper),
                       upper=self._upper | self._to_denominator(other._lower),
                       precision=min(self.precision, other.precision))
        if isinstance(other, Iterable):
            if not isinstance(other, BoundedSet):
                other = BoundedSet(other)
            return BoundedSet(other,
                              lower={self.min()} | other._lower,
                              upper={self.max()} | other._upper,
                              precision=min(self.precision, other.precision))
        return NotImplemented

    def to_bounded_set(self) -> BoundedSet:
        return BoundedSet((self.numerator / i for i in range(self.lower, self.upper + 1)),
                          lower=self.min(),
                          upper=self.max(),
                          precision=self.precision)

    def next_closest(self, value: _RealT) -> _RealT:
        return self.numerator / self._denominator_interval.next_closest(self.numerator / value)

    def next_smallest(self, value: _RealT) -> _RealT:
        return self.numerator / self._denominator_interval.next_largest(self.numerator / value)

    def next_largest(self, value: _RealT) -> _RealT:
        return self.numerator / self._denominator_interval.next_smallest(self.numerator / value)

    def min(self) -> _RealT:
        return self.numerator / self.upper

    def max(self) -> _RealT:
        return self.numerator / self.lower


@dataclass
class BoundedSet(frozenset, Domain[_RealT], Bounded[_RealT]):
    """A finite set of numbers with lower and upper bounds.

    Examples
    --------
    >>> domain = BoundedSet([1, 1.0, 3, -2.3, 7], lower=-4, upper=5)
    >>> domain  # behaves like frozenset
    BoundedSet({1, 3, -2.3, 7}, lower=-4, upper=5)
    >>> 3 in domain
    True
    >>> -1.7 in domain
    False
    >>> 7 in domain  # out of bounds
    False
    >>> domain.next_largest(2.7)
    3

    """
    iterable: InitVar[Iterable[_RealT]] = ...
    lower: _BoundT[_RealT] = Bound[_RealT](default=-inf, integral=False)
    upper: _BoundT[_RealT] = Bound[_RealT](default=+inf, integral=False)
    precision: int = field(default=16, repr=False)

    def __repr__(self) -> str:
        r = super().__repr__()
        r = r.replace(')', f', lower={self.lower}, upper={self.upper})', 1)
        return r

    def __and__(self, other: object) -> _DomainT:
        cls = type(self)
        if isinstance(other, Domain) and not self:
            return other
        if isinstance(other, Interval):
            return cls(self,
                       lower=self._lower | other._lower,
                       upper=self._upper | other._upper,
                       precision=min(self.precision, other.precision))
        if isinstance(other, BoundedSet):
            return cls(frozenset.intersection(self, other),
                       lower=self._lower | other._lower,
                       upper=self._upper | other._upper,
                       precision=min(self.precision, other.precision))
        if isinstance(other, Iterable):
            return cls(frozenset.intersection(self, other),
                       lower=self._lower,
                       upper=self._upper,
                       precision=min(self.precision, other.precision))
        return NotImplemented

    def __contains__(self, value: object) -> bool:
        return (isinstance(value, numbers.Real)
                and self.round(self.lower - value) <= 0 <= self.round(self.upper - value)
                and self.round(value) in self._rounded)

    @cached_property
    def _rounded(self) -> Set[_RealT]:
        return {self.round(item) for item in self}

    def min(self) -> _RealT:
        return min((item for item in self if item in self), default=nan)

    def max(self) -> _RealT:
        return max((item for item in self if item in self), default=nan)

    def next_closest(self, value: _RealT) -> _RealT:
        return next_closest(self, value, self.precision)

    def next_smallest(self, value: _RealT) -> _RealT:
        return next_smallest(self, value, self.precision)

    def next_largest(self, value: _RealT) -> _RealT:
        return next_largest(self, value, self.precision)
